#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import threading
import weakref
import itertools
from collections import deque, defaultdict

import numpy as np

from ...operands import Fetch
from ...graph import DirectedGraph
from ...compat import futures
from ..core import build_mode
from ..expressions.merge.concatenate import TensorConcatenate


class Executor(object):
    _op_runners = {}

    def __init__(self, engine=None, storage=None, prefetch=False):
        self._engine = engine
        self._chunk_result = storage if storage is not None else dict()
        self._prefetch = prefetch

        # only record the executed tensor
        self.tensor_to_tiled = weakref.WeakKeyDictionary()
        self.tensor_key_to_chunk_keys = dict()
        # executed key to ref counts
        self.key_to_ref_counts = defaultdict(lambda: 0)

    @property
    def chunk_result(self):
        return self._chunk_result

    def _preprocess(self, graph, keys):
        from .optimizes.core import Optimizer

        Optimizer(graph, self._engine).optimize(keys=keys)
        return graph

    def handle(self, chunk, results):
        try:
            cls = type(chunk.op)

            return self._op_runners[cls](results, chunk)
        except KeyError:
            for cls in self._op_runners.keys():
                if isinstance(chunk.op, cls):
                    self._op_runners[type(chunk.op)] = self._op_runners[cls]
                    return self._op_runners[cls](results, chunk)

            raise KeyError('No handler found for op: %s' % chunk.op)

    def execute_graph(self, graph, keys, n_parallel=None, n_thread=None, show_progress=False,
                      mock=False, sparse_mock_percent=1.0):
        with build_mode():
            optimized_graph = self._preprocess(graph, keys)

        return execute_graph(optimized_graph, keys, self, n_parallel=n_parallel or n_thread,
                             show_progress=show_progress, mock=mock,
                             sparse_mock_percent=sparse_mock_percent,
                             prefetch=self._prefetch, retval=True)

    def execute_tensor(self, tensor, n_parallel=None, n_thread=None, concat=False,
                       show_progress=False, mock=False, sparse_mock_percent=1.0):
        if concat:
            # only for tests
            tensor.tiles()
            if len(tensor.chunks) > 1:
                op = TensorConcatenate(dtype=tensor.op.dtype)
                chunk = TensorConcatenate(dtype=op.dtype).new_chunk(tensor.chunks, tensor.shape)
                tensor = op.new_tensor([tensor], tensor.shape, chunks=[chunk],
                                       nsplits=[(s,) for s in tensor.shape])

        graph = tensor.build_graph(cls=DirectedGraph, tiled=True)

        return self.execute_graph(graph, [c.key for c in tensor.chunks],
                                  n_parallel=n_parallel or n_thread,
                                  show_progress=show_progress, mock=mock,
                                  sparse_mock_percent=sparse_mock_percent)

    def execute_tensors(self, tensors, n_parallel=None, n_thread=None,
                        show_progress=False, mock=False, sparse_mock_percent=1.0):
        graph = DirectedGraph()

        result_keys = []
        to_release_keys = []
        concat_keys = []
        for tensor in tensors:
            tensor.tiles()
            chunk_keys = [c.key for c in tensor.chunks]
            self.tensor_key_to_chunk_keys[(tensor.key, tensor.id)] = chunk_keys
            result_keys.extend(chunk_keys)

            if len(tensor.chunks) > 1:
                # if chunks more than 1, we concatenate them into 1
                op = TensorConcatenate(dtype=tensor.op.dtype)
                chunk = TensorConcatenate(dtype=op.dtype).new_chunk(tensor.chunks, tensor.shape)
                result_keys.append(chunk.key)
                # the concatenated key
                concat_keys.append(chunk.key)
                # after return the data to user, we release the reference
                to_release_keys.append(chunk.key)
                tensor = op.new_tensor([tensor], tensor.shape, chunks=[chunk],
                                       nsplits=[(s,) for s in tensor.shape])
            else:
                concat_keys.append(tensor.chunks[0].key)

            tensor.build_graph(graph=graph, tiled=True)

        self.execute_graph(graph, result_keys, n_parallel=n_parallel or n_thread,
                           show_progress=show_progress, mock=mock,
                           sparse_mock_percent=sparse_mock_percent)

        results = self._chunk_result
        try:
            return [results[k] for k in concat_keys]
        finally:
            for k in to_release_keys:
                del results[k]

    def decref(self, *keys):
        for key in keys:
            for chunk_key in self.tensor_key_to_chunk_keys.get(key, []):
                if chunk_key in self.chunk_result:
                    del self.chunk_result[chunk_key]


def execute_chunk(chunk, executor=None,
                  ref_counts=None, chunk_result=None,
                  finishes=None, visited=None, q=None,
                  lock=None, semaphore=None, has_error=None,
                  preds=None, succs=None, mock=False, sparse_mock_percent=1.0):
    try:
        with lock:
            if (chunk.key, chunk.id) in visited:
                return
            visited.add((chunk.key, chunk.id))
            finished = finishes.get(chunk.key)
        if not finished:
            if not mock:
                # do real execution
                if chunk.key not in chunk_result:
                    executor.handle(chunk, chunk_result)
            else:
                percent = sparse_mock_percent if chunk.op.sparse else 1.0
                # we put the estimated size of data into the chunk_result
                chunk_result[chunk.key] = np.prod(chunk.shape) * chunk.dtype.itemsize * percent
            with lock:
                for output in chunk.op.outputs:
                    finishes[output.key] = True

        for pred_key in preds[chunk.key]:
            with lock:
                if pred_key not in ref_counts:
                    continue
                ref_counts[pred_key] -= 1
                if ref_counts[pred_key] == 0:
                    del chunk_result[pred_key]

        for succ in succs[chunk.key, chunk.id]:
            with lock:
                if (succ.key, succ.id) in visited:
                    continue
                if len(preds[succ.key]) == 0 or \
                        all(finishes.get(k, False) for k in preds[succ.key]):
                    q.insert(0, succ)
    except Exception:
        has_error.set()
        raise
    finally:
        semaphore.release()


def _order_starts(graph):
    visited = set()
    starts = deque(graph.iter_indep())
    stack = deque([starts.popleft()])

    while stack:
        node = stack.popleft()
        if node not in visited:
            preds = graph.predecessors(node)
            if not preds or all(pred in visited for pred in preds):
                if len(preds) == 0:
                    yield node
                visited.add(node)
                stack.extend(n for n in graph[node] if n not in visited)
            else:
                stack.appendleft(node)
                stack.extendleft(reversed(list(n for n in graph.predecessors(node)
                                               if n not in visited)))
        if not stack and starts:
            stack.appendleft(starts.popleft())


def execute_graph(graph, keys, executor, n_parallel=None, show_progress=False,
                  mock=False, sparse_mock_percent=1.0, prefetch=False, retval=True):
    pool_executor = futures.ThreadPoolExecutor(n_parallel or 1)
    prefetch_executor = futures.ThreadPoolExecutor(n_parallel or 1) if prefetch else None
    chunk_result = executor.chunk_result
    q = list()
    lock = threading.Lock()
    semaphore = threading.Semaphore(pool_executor._max_workers)
    has_error = threading.Event()

    preds = dict()
    preds.update(
        dict((t.key, [i.key for i in v]) for t, v in graph.iter_predecessor_items())
    )
    succs = dict()
    succs.update(
        dict(((t.key, t.id), v) for t, v in graph.iter_successor_items())
    )

    starts = list(_order_starts(graph)) if len(graph) > 0 else list()
    assert len(starts) == sum(1 for _ in graph.iter_indep())
    q.extend(starts)
    finishes = dict()
    visited = set()
    fs = dict()
    ref_counts = dict()
    key_set = set(keys)
    for chunk in graph:
        if chunk.key not in key_set:
            ref_counts[chunk.key] = ref_counts.get(chunk.key, 0) + len(graph[chunk])
    node_keys_set = {n.key for n in graph}
    count = itertools.count(0)
    maximum_usage = 0

    def fetch(chunk):
        with lock:
            try:
                to_fetch_chunk = next(c for c in list(succs.get((chunk.key, chunk.id), []))
                                      if all(i in finishes for i in preds[c.key] if i != chunk.key))
            except StopIteration:
                if len(q) > 0:
                    to_fetch_chunk = q[0]
                else:
                    return

        [chunk_result.get(k) for k in preds[to_fetch_chunk.key]]

    def submit_to_execute():
        if show_progress:
            c = next(count)
            if c % 30 == 0 or c >= len(graph):
                print('[{0}] {1:.2f}% percent of graph has been submitted'.format(str(datetime.datetime.now()), float(c) * 100 / len(graph)))

        with lock:
            if len(q) == 0:
                semaphore.release()
                return
            chunk = q.pop(0)

        if prefetch:
            prefetch_executor.submit(fetch, chunk)
        future = pool_executor.submit(execute_chunk, chunk, executor=executor,
                                      ref_counts=ref_counts, chunk_result=chunk_result,
                                      finishes=finishes, visited=visited, q=q,
                                      lock=lock, semaphore=semaphore, has_error=has_error,
                                      preds=preds, succs=succs,
                                      mock=mock, sparse_mock_percent=sparse_mock_percent)
        fs[chunk.key] = future

    while len(node_keys_set - set(finishes.keys())) > 0:
        if has_error.is_set():
            break
        semaphore.acquire()
        if mock:
            maximum_usage = max(maximum_usage, np.sum(list(chunk_result.values())))
            # if maximum_usage > 40 * (1024 ** 3):
            #     raise RuntimeError('memory exceed')
        submit_to_execute()

    if mock:
        [f.result() for f in fs.values()]
        return maximum_usage

    [f.result() for f in fs.values()]
    if retval:
        return [chunk_result[key] for key in keys]


def ignore(*_):
    pass
Executor._op_runners[Fetch] = ignore


def register(op, handler):
    Executor._op_runners[op] = handler


from .datasource import register_data_source_handler
from .random import register_random_handler
from .base import register_basic_handler
from .arithmetic import register_arithmetic_handler
from .indexing import register_indexing_handler
from .reduction import register_reduction_handler
from .merge import register_merge_handler
from .fft import register_fft_handler
from .linalg import register_linalg_handler

NUMEXPR_INSTALLED = False
try:
    import numexpr
    NUMEXPR_INSTALLED = True
    from .ne import register_numexpr_handler
    register_numexpr_handler()
except ImportError:
    pass

CP_INSTALLED = False
try:
    import cupy
    CP_INSTALLED = True
    from .cp import register_cp_handler
    register_cp_handler()
except ImportError:
    pass

register_data_source_handler()
register_random_handler()
register_basic_handler()
register_arithmetic_handler()
register_indexing_handler()
register_reduction_handler()
register_merge_handler()
register_fft_handler()
register_linalg_handler()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorUnaryOp


class TensorArcsin(operands.Arcsin, TensorUnaryOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorArcsin, self).__init__(_casting=casting, _err=err,
                                           _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x):
        return x.issparse()


@infer_dtype(np.arcsin)
def arcsin(x, out=None, where=None, **kwargs):
    """
    Inverse sine, element-wise.

    Parameters
    ----------
    x : array_like
        `y`-coordinate on the unit circle.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    angle : Tensor
        The inverse sine of each element in `x`, in radians and in the
        closed interval ``[-pi/2, pi/2]``.  If `x` is a scalar, a scalar
        is returned, otherwise a tensor.

    See Also
    --------
    sin, cos, arccos, tan, arctan, arctan2, emath.arcsin

    Notes
    -----
    `arcsin` is a multivalued function: for each `x` there are infinitely
    many numbers `z` such that :math:`sin(z) = x`.  The convention is to
    return the angle `z` whose real part lies in [-pi/2, pi/2].

    For real-valued input data types, *arcsin* always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    For complex-valued input, `arcsin` is a complex analytic function that
    has, by convention, the branch cuts [-inf, -1] and [1, inf]  and is
    continuous from above on the former and from below on the latter.

    The inverse sine is also known as `asin` or sin^{-1}.

    References
    ----------
    Abramowitz, M. and Stegun, I. A., *Handbook of Mathematical Functions*,
    10th printing, New York: Dover, 1964, pp. 79ff.
    http://www.math.sfu.ca/~cbm/aands/

    Examples
    --------
    >>> import mars.tensor as mt
    >>> mt.arcsin(1).execute()     # pi/2
    1.5707963267948966
    >>> mt.arcsin(-1).execute()    # -pi/2
    -1.5707963267948966
    >>> mt.arcsin(0).execute()
    0.0
    """
    op = TensorArcsin(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .array import tensor, array, asarray, ArrayDataSource, CSRMatrixDataSource
from .scalar import scalar, Scalar
from .empty import empty, empty_like, TensorEmpty, TensorEmptyLike
from .ones import ones, ones_like, TensorOnes, TensorOnesLike
from .zeros import zeros, zeros_like, TensorZeros, TensorZerosLike
from .full import full, TensorFull
from .arange import arange, TensorArange
from .diag import diag, TensorDiag
from .diagflat import diagflat
from .eye import eye, TensorEye
from .identity import identity
from .linspace import linspace, TensorLinspace
from .meshgrid import meshgrid
from .indices import indices, TensorIndices
from .tri import triu, tril, TensorTriu, TensorTril
from .fromdense import fromdense, DenseToSparse
from .fromsparse import fromsparse, SparseToDense
from .core import TensorFetchChunk



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bokeh.models import ColumnDataSource
from bokeh.plotting import figure

from .server import register_ui_handler, get_jinja_env
from ..scheduler.utils import OperandState
from ..utils import to_str
from ..actors import new_client
from .server import MarsWebAPI


_actor_client = new_client()
_jinja_env = get_jinja_env()

_base_palette = ['#e0f9ff', '#81d4dd', '#ff9e3b', '#399e34', '#2679b2', '#b3de8e',
                 '#e01f27', '#b3b3cc', '#f0f0f5']


def task_list(doc, web_api):
    sessions = web_api.get_tasks_info()

    doc.title = 'Mars UI'
    doc.template_variables['sessions'] = sessions
    doc.template = _jinja_env.get_template('task_list.html')


def task_operand(doc, cluster_info, session_id, task_id):
    session_name = session_id
    states = list(OperandState.__members__.values())

    ops, stats, progress = cluster_info.get_task_detail(session_id, task_id)
    source = ColumnDataSource(stats)
    cols = list(stats)[1:]
    p = figure(y_range=[''] + ops, plot_height=500, plot_width=800, x_range=(0, 100),
               title="Task Progresses")
    p.hbar_stack(cols, y='ops', height=0.9, color=_base_palette[0:len(states)],
                 source=source, legend=['\0' + s for s in cols])

    p.legend.background_fill_alpha = 0.5
    p.legend.location = 'bottom_left'
    p.legend.orientation = "horizontal"
    p.legend.label_text_font_size = '10px'

    p.y_range.range_padding = 0.1
    p.ygrid.grid_line_color = None
    p.yaxis.major_tick_line_color = None
    p.yaxis.minor_tick_line_color = None
    p.axis.minor_tick_line_color = None
    p.outline_line_color = None

    doc.add_root(p)

    doc.title = 'Mars UI'
    doc.template_variables['session_id'] = session_id
    doc.template_variables['session_name'] = session_name
    doc.template_variables['task_id'] = task_id
    doc.template_variables['progress'] = progress
    doc.template = _jinja_env.get_template('task_operands.html')


def _route(scheduler_ip, doc):
    web_api = MarsWebAPI(scheduler_ip)
    session_id = doc.session_context.request.arguments.get('session_id')
    task_id = doc.session_context.request.arguments.get('task_id')
    if session_id and task_id:
        task_operand(doc, web_api, to_str(session_id[0]), to_str(task_id[0]))
    else:
        task_list(doc, web_api)


register_ui_handler('/task', _route)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorUnaryOp


class TensorSinc(operands.Sinc, TensorUnaryOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorSinc, self).__init__(_casting=casting, _err=err,
                                         _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x):
        return False


@infer_dtype(np.sinc)
def sinc(x, **kwargs):
    """
    Return the sinc function.

    The sinc function is :math:`\\sin(\\pi x)/(\\pi x)`.

    Parameters
    ----------
    x : Tensor
        Tensor (possibly multi-dimensional) of values for which to to
        calculate ``sinc(x)``.

    Returns
    -------
    out : Tensor
        ``sinc(x)``, which has the same shape as the input.

    Notes
    -----
    ``sinc(0)`` is the limit value 1.

    The name sinc is short for "sine cardinal" or "sinus cardinalis".

    The sinc function is used in various signal processing applications,
    including in anti-aliasing, in the construction of a Lanczos resampling
    filter, and in interpolation.

    For bandlimited interpolation of discrete-time signals, the ideal
    interpolation kernel is proportional to the sinc function.

    References
    ----------
    .. [1] Weisstein, Eric W. "Sinc Function." From MathWorld--A Wolfram Web
           Resource. http://mathworld.wolfram.com/SincFunction.html
    .. [2] Wikipedia, "Sinc function",
           http://en.wikipedia.org/wiki/Sinc_function

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.linspace(-4, 4, 41)
    >>> mt.sinc(x).execute()
    array([ -3.89804309e-17,  -4.92362781e-02,  -8.40918587e-02,
            -8.90384387e-02,  -5.84680802e-02,   3.89804309e-17,
             6.68206631e-02,   1.16434881e-01,   1.26137788e-01,
             8.50444803e-02,  -3.89804309e-17,  -1.03943254e-01,
            -1.89206682e-01,  -2.16236208e-01,  -1.55914881e-01,
             3.89804309e-17,   2.33872321e-01,   5.04551152e-01,
             7.56826729e-01,   9.35489284e-01,   1.00000000e+00,
             9.35489284e-01,   7.56826729e-01,   5.04551152e-01,
             2.33872321e-01,   3.89804309e-17,  -1.55914881e-01,
            -2.16236208e-01,  -1.89206682e-01,  -1.03943254e-01,
            -3.89804309e-17,   8.50444803e-02,   1.26137788e-01,
             1.16434881e-01,   6.68206631e-02,   3.89804309e-17,
            -5.84680802e-02,  -8.90384387e-02,  -8.40918587e-02,
            -4.92362781e-02,  -3.89804309e-17])

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(x.execute(), np.sinc(x).execute())
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> plt.title("Sinc Function")
    <matplotlib.text.Text object at 0x...>
    >>> plt.ylabel("Amplitude")
    <matplotlib.text.Text object at 0x...>
    >>> plt.xlabel("X")
    <matplotlib.text.Text object at 0x...>
    >>> plt.show()
    """
    op = TensorSinc(**kwargs)
    return op(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorArgReduction


class TensorArgminChunk(operands.ArgminChunk, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None,
                 combine_size=None, offset=None, total_shape=None,**kw):
        super(TensorArgminChunk, self).__init__(_axis=axis, _dtype=dtype,
                                                _keepdims=keepdims, _combine_size=combine_size,
                                                _offset=offset, _total_shape=total_shape, **kw)


class TensorArgminCombine(operands.ArgminCombine, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None, combine_size=None, **kw):
        super(TensorArgminCombine, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                                  _combine_size=combine_size, **kw)


class TensorArgmin(operands.Argmin, TensorArgReduction):
    def __init__(self, axis=None, dtype=np.dtype(int), keepdims=None, combine_size=None, **kw):
        super(TensorArgmin, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                           _combine_size=combine_size, **kw)

    @staticmethod
    def _get_op_types():
        return TensorArgminChunk, TensorArgmin, TensorArgminCombine


def argmin(a, axis=None, out=None, keepdims=None, combine_size=None):
    """
    Returns the indices of the minimum values along an axis.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axis : int, optional
        By default, the index is into the flattened tensor, otherwise
        along the specified axis.
    out : Tensor, optional
        If provided, the result will be inserted into this tensor. It should
        be of the appropriate shape and dtype.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the input tensor.

        If the default value is passed, then `keepdims` will not be
        passed through to the `mean` method of sub-classes of
        `Tensor`, however any non-default value will be.  If the
        sub-classes `sum` method does not implement `keepdims` any
        exceptions will be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    index_array : Tensor of ints
        Tensor of indices into the tensor. It has the same shape as `a.shape`
        with the dimension along `axis` removed.

    See Also
    --------
    Tensor.argmin, argmax
    amin : The minimum value along a given axis.
    unravel_index : Convert a flat index into an index tuple.

    Notes
    -----
    In case of multiple occurrences of the minimum values, the indices
    corresponding to the first occurrence are returned.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.session import new_session

    >>> sess = new_session().as_default()

    >>> a = mt.arange(6).reshape(2,3)
    >>> a.execute()
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> mt.argmin(a).execute()
    0
    >>> mt.argmin(a, axis=0).execute()
    array([0, 0, 0])
    >>> mt.argmin(a, axis=1).execute()
    array([0, 0])

    Indices of the minimum elements of a N-dimensional tensor:

    >>> ind = mt.unravel_index(mt.argmin(a, axis=None), a.shape)
    >>> sess.run(ind)
    (0, 0)
    >>> a[ind]  # TODO(jisheng): accomplish when fancy index on tensor is supported

    >>> b = mt.arange(6)
    >>> b[4] = 0
    >>> b.execute()
    array([0, 1, 2, 3, 0, 5])
    >>> mt.argmin(b).execute()  # Only the first occurrence is returned.
    0

    """
    op = TensorArgmin(axis=axis, dtype=np.dtype(int), keepdims=keepdims, combine_size=combine_size)
    return op(a, out=out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

dirpath = os.path.dirname
parent_path = dirpath(dirpath(dirpath(__file__)))

sys.path.insert(0, dirpath(parent_path))

from mars.serialize.protos.operand_pb2 import OperandDef


with open(os.path.join(parent_path, 'opcodes.py'), 'w') as f:
    f.write('# Generated automatically.  DO NOT EDIT!\n')
    for val, desc in OperandDef.OperandType.DESCRIPTOR.values_by_number.items():
        f.write('{0} = {1!r}\n'.format(desc.name, val))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ....compat import izip
from ..utils import decide_chunks
from ..core import TensorOperandMixin


class TSQR(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def _is_svd(cls):
        return False

    @staticmethod
    def _get_obj_attr(x, attr):
        return getattr(x, attr) if x is not None else None

    @classmethod
    def tile(cls, op):
        from ..merge.concatenate import TensorConcatenate
        from ..indexing.slice import TensorSlice
        from .dot import TensorDot
        from .qr import TensorQR
        from .svd import TensorSVD

        calc_svd = cls._is_svd()
        get_obj_attr = cls._get_obj_attr

        a = op.input

        tinyq, tinyr = np.linalg.qr(np.ones((1, 1), dtype=a.dtype))
        q_dtype, r_dtype = tinyq.dtype, tinyr.dtype

        if a.chunk_shape[1] != 1:
            new_chunks = decide_chunks(a.shape, {1: a.shape[1]}, a.dtype.itemsize)
            a = a.rechunk(new_chunks).single_tiles()

        # stage 1, map phase
        stage1_q_chunks, stage1_r_chunks = stage1_chunks = [[], []]  # Q and R chunks
        for c in a.chunks:
            x, y = c.shape
            q_shape = c.shape if x > y else (x, x)
            r_shape = c.shape if x < y else (y, y)
            qr_op = TensorQR()
            qr_chunks = qr_op.new_chunks([c], [q_shape, r_shape], index=c.index,
                                         kws=[{'side': 'q', 'dtype': q_dtype},
                                              {'side': 'r', 'dtype': r_dtype}])
            stage1_chunks[0].append(qr_chunks[0])
            stage1_chunks[1].append(qr_chunks[1])

        # stage 2, reduce phase
        # concatenate all r chunks into one
        shape = (sum(c.shape[0] for c in stage1_r_chunks), stage1_r_chunks[0].shape[1])
        concat_op = TensorConcatenate(axis=0, dtype=stage1_r_chunks[0].dtype)
        concat_r_chunk = concat_op.new_chunk(stage1_r_chunks, shape, index=(0, 0))
        qr_op = TensorQR()
        qr_shapes = concat_r_chunk.shape, (concat_r_chunk.shape[1],) * 2
        qr_chunks = qr_op.new_chunks([concat_r_chunk], qr_shapes, index=concat_r_chunk.index,
                                     kws=[{'side': 'q', 'dtype': q_dtype},
                                          {'side': 'r', 'dtype': r_dtype}])
        stage2_q_chunk, stage2_r_chunk = qr_chunks

        # stage 3, map phase
        # split stage2_q_chunk into the same size as stage1_q_chunks
        q_splits = np.cumsum([c.shape[1] for c in stage1_q_chunks])
        q_slices = [slice(q_splits[i]) if i == 0 else slice(q_splits[i-1], q_splits[i])
                    for i in range(len(q_splits))]
        stage2_q_chunks = []
        for c, s in zip(stage1_q_chunks, q_slices):
            slice_op = TensorSlice(slices=[s], dtype=c.dtype)
            stage2_q_chunks.append(slice_op.new_chunk([stage2_q_chunk], c.shape, index=c.index))
        stage3_q_chunks = []
        for c1, c2 in izip(stage1_q_chunks, stage2_q_chunks):
            dot_op = TensorDot(dtype=q_dtype)
            shape = (c1.shape[0], c2.shape[1])
            stage3_q_chunks.append(dot_op.new_chunk([c1, c2], shape, index=c1.index))

        if not calc_svd:
            q, r = op.outputs
            new_op = op.copy()
            q_nsplits = ((c.shape[0] for c in stage3_q_chunks), (1,))
            r_nsplits = ((stage2_r_chunk.shape[0],), (stage2_r_chunk.shape[1],))
            kws = [
                # Q
                {'chunks': stage3_q_chunks, 'nsplits': q_nsplits, 'dtype': get_obj_attr(q, 'dtype')},
                # R, calculate from stage2
                {'chunks': [stage2_r_chunk], 'nsplits': r_nsplits, 'dtype': get_obj_attr(r, 'dtype')}
            ]
            return new_op.new_tensors(op.inputs, [get_obj_attr(q, 'shape'), get_obj_attr(r, 'shape')], kws=kws)
        else:
            U, s, V = op.outputs
            U_dtype, s_dtype, V_dtype = \
                get_obj_attr(U, 'dtype'), get_obj_attr(s, 'dtype'), get_obj_attr(V, 'dtype')
            U_shape, s_shape, V_shape = \
                get_obj_attr(U, 'shape'), get_obj_attr(s, 'shape'), get_obj_attr(V, 'shape')

            svd_op = TensorSVD()
            u_shape = stage2_r_chunk.shape
            s_shape = (stage2_r_chunk.shape[1],)
            v_shape = (stage2_r_chunk.shape[1],) * 2
            stage2_usv_chunks = svd_op.new_chunks([stage2_r_chunk], [u_shape, s_shape, v_shape],
                                                  index=stage2_r_chunk.index,
                                                  kws=[{'side': 'U', 'dtype': U_dtype},
                                                       {'side': 's', 'dtype': s_dtype},
                                                       {'side': 'V', 'dtype': V_dtype}])
            stage2_u_chunk, stage2_s_chunk, stage2_v_chunk = stage2_usv_chunks

            # stage 4, U = Q @ u
            stage4_u_chunks = []
            if U is not None:  # U is not garbage collected
                for c1 in stage3_q_chunks:
                    dot_op = TensorDot(dtype=U_dtype)
                    shape = (c1.shape[0], stage2_u_chunk.shape[1])
                    stage4_u_chunks.append(dot_op.new_chunk([c1, stage2_u_chunk], shape,
                                                            index=c1.index))

            new_op = op.copy()
            u_nsplits = ((c.shape[0] for c in stage4_u_chunks), (1,))
            s_nsplits = ((stage2_s_chunk.shape[0],),)
            v_nsplits = ((stage2_v_chunk.shape[0],), (stage2_v_chunk.shape[1],))
            kws = [
                {'chunks': stage4_u_chunks, 'nsplits': u_nsplits, 'dtype': U_dtype},   # U
                {'chunks': [stage2_s_chunk], 'nsplits': s_nsplits, 'dtype': s_dtype},  # s
                {'chunks': [stage2_v_chunk], 'nsplits': v_nsplits, 'dtype': V_dtype},  # V
            ]
            return new_op.new_tensors(op.inputs, [U_shape, s_shape, V_shape], kws=kws)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def allclose(a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
    """
    Returns True if two tensors are element-wise equal within a tolerance.

    The tolerance values are positive, typically very small numbers.  The
    relative difference (`rtol` * abs(`b`)) and the absolute difference
    `atol` are added together to compare against the absolute difference
    between `a` and `b`.

    If either array contains one or more NaNs, False is returned.
    Infs are treated as equal if they are in the same place and of the same
    sign in both tensors.

    Parameters
    ----------
    a, b : array_like
        Input tensors to compare.
    rtol : float
        The relative tolerance parameter (see Notes).
    atol : float
        The absolute tolerance parameter (see Notes).
    equal_nan : bool
        Whether to compare NaN's as equal.  If True, NaN's in `a` will be
        considered equal to NaN's in `b` in the output tensor.

    Returns
    -------
    allclose : bool
        Returns True if the two tensors are equal within the given
        tolerance; False otherwise.

    See Also
    --------
    isclose, all, any, equal

    Notes
    -----
    If the following equation is element-wise True, then allclose returns
    True.

     absolute(`a` - `b`) <= (`atol` + `rtol` * absolute(`b`))

    The above equation is not symmetric in `a` and `b`, so that
    ``allclose(a, b)`` might be different from ``allclose(b, a)`` in
    some rare cases.

    The comparison of `a` and `b` uses standard broadcasting, which
    means that `a` and `b` need not have the same shape in order for
    ``allclose(a, b)`` to evaluate to True.  The same is true for
    `equal` but not `array_equal`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.allclose([1e10,1e-7], [1.00001e10,1e-8]).execute()
    False
    >>> mt.allclose([1e10,1e-8], [1.00001e10,1e-9]).execute()
    True
    >>> mt.allclose([1e10,1e-8], [1.0001e10,1e-9]).execute()
    False
    >>> mt.allclose([1.0, mt.nan], [1.0, mt.nan]).execute()
    False
    >>> mt.allclose([1.0, mt.nan], [1.0, mt.nan], equal_nan=True).execute()
    True

    """
    from ..arithmetic.isclose import isclose
    from .all import all

    return all(isclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorUnaryOp


class TensorExp(operands.Exp, TensorUnaryOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorExp, self).__init__(_casting=casting, _err=err,
                                        _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x):
        return False


@infer_dtype(np.exp)
def exp(x, out=None, where=None, **kwargs):
    """
    Calculate the exponential of all elements in the input tensor.

    Parameters
    ----------
    x : array_like
        Input values.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs
        For other keyword-only arguments, see the
        :ref:`ufunc docs <ufuncs.kwargs>`.

    Returns
    -------
    out : Tensor
        Output tensor, element-wise exponential of `x`.

    See Also
    --------
    expm1 : Calculate ``exp(x) - 1`` for all elements in the array.
    exp2  : Calculate ``2**x`` for all elements in the array.

    Notes
    -----
    The irrational number ``e`` is also known as Euler's number.  It is
    approximately 2.718281, and is the base of the natural logarithm,
    ``ln`` (this means that, if :math:`x = \ln y = \log_e y`,
    then :math:`e^x = y`. For real input, ``exp(x)`` is always positive.

    For complex arguments, ``x = a + ib``, we can write
    :math:`e^x = e^a e^{ib}`.  The first term, :math:`e^a`, is already
    known (it is the real argument, described above).  The second term,
    :math:`e^{ib}`, is :math:`\cos b + i \sin b`, a function with
    magnitude 1 and a periodic phase.

    References
    ----------
    .. [1] Wikipedia, "Exponential function",
           http://en.wikipedia.org/wiki/Exponential_function
    .. [2] M. Abramovitz and I. A. Stegun, "Handbook of Mathematical Functions
           with Formulas, Graphs, and Mathematical Tables," Dover, 1964, p. 69,
           http://www.math.sfu.ca/~cbm/aands/page_69.htm

    Examples
    --------
    Plot the magnitude and phase of ``exp(x)`` in the complex plane:

    >>> import mars.tensor as mt
    >>> import matplotlib.pyplot as plt

    >>> x = mt.linspace(-2*mt.pi, 2*mt.pi, 100)
    >>> xx = x + 1j * x[:, mt.newaxis] # a + ib over complex plane
    >>> out = mt.exp(xx)

    >>> plt.subplot(121)
    >>> plt.imshow(mt.abs(out).execute(),
    ...            extent=[-2*mt.pi, 2*mt.pi, -2*mt.pi, 2*mt.pi], cmap='gray')
    >>> plt.title('Magnitude of exp(x)')

    >>> plt.subplot(122)
    >>> plt.imshow(mt.angle(out).execute(),
    ...            extent=[-2*mt.pi, 2*mt.pi, -2*mt.pi, 2*mt.pi], cmap='hsv')
    >>> plt.title('Phase (angle) of exp(x)')
    >>> plt.show()
    """
    op = TensorExp(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorPareto(operands.Pareto, TensorRandomOperandMixin):
    __slots__ = '_a', '_size'
    _input_fields_ = ['_a']

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorPareto, self).__init__(_size=size, _state=state, _dtype=dtype,
                                           _gpu=gpu, **kw)

    def __call__(self, a, chunks=None):
        return self.new_tensor([a], None, raw_chunks=chunks)


def pareto(random_state, a, size=None, chunks=None, gpu=None, **kw):
    """
    Draw samples from a Pareto II or Lomax distribution with
    specified shape.

    The Lomax or Pareto II distribution is a shifted Pareto
    distribution. The classical Pareto distribution can be
    obtained from the Lomax distribution by adding 1 and
    multiplying by the scale parameter ``m`` (see Notes).  The
    smallest value of the Lomax distribution is zero while for the
    classical Pareto distribution it is ``mu``, where the standard
    Pareto distribution has location ``mu = 1``.  Lomax can also
    be considered as a simplified version of the Generalized
    Pareto distribution (available in SciPy), with the scale set
    to one and the location set to zero.

    The Pareto distribution must be greater than zero, and is
    unbounded above.  It is also known as the "80-20 rule".  In
    this distribution, 80 percent of the weights are in the lowest
    20 percent of the range, while the other 20 percent fill the
    remaining 80 percent of the range.

    Parameters
    ----------
    a : float or array_like of floats
        Shape of the distribution. Should be greater than zero.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``a`` is a scalar.  Otherwise,
        ``mt.array(a).size`` samples are drawn.
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized Pareto distribution.

    See Also
    --------
    scipy.stats.lomax : probability density function, distribution or
        cumulative density function, etc.
    scipy.stats.genpareto : probability density function, distribution or
        cumulative density function, etc.

    Notes
    -----
    The probability density for the Pareto distribution is

    .. math:: p(x) = \frac{am^a}{x^{a+1}}

    where :math:`a` is the shape and :math:`m` the scale.

    The Pareto distribution, named after the Italian economist
    Vilfredo Pareto, is a power law probability distribution
    useful in many real world problems.  Outside the field of
    economics it is generally referred to as the Bradford
    distribution. Pareto developed the distribution to describe
    the distribution of wealth in an economy.  It has also found
    use in insurance, web page access statistics, oil field sizes,
    and many other problems, including the download frequency for
    projects in Sourceforge [1]_.  It is one of the so-called
    "fat-tailed" distributions.


    References
    ----------
    .. [1] Francis Hunt and Paul Johnson, On the Pareto Distribution of
           Sourceforge projects.
    .. [2] Pareto, V. (1896). Course of Political Economy. Lausanne.
    .. [3] Reiss, R.D., Thomas, M.(2001), Statistical Analysis of Extreme
           Values, Birkhauser Verlag, Basel, pp 23-30.
    .. [4] Wikipedia, "Pareto distribution",
           http://en.wikipedia.org/wiki/Pareto_distribution

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> a, m = 3., 2.  # shape and mode
    >>> s = (mt.random.pareto(a, 1000) + 1) * m

    Display the histogram of the samples, along with the probability
    density function:

    >>> import matplotlib.pyplot as plt
    >>> count, bins, _ = plt.hist(s.execute(), 100, normed=True)
    >>> fit = a*m**a / bins**(a+1)
    >>> plt.plot(bins, max(count)*fit/max(fit), linewidth=2, color='r')
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().pareto(
            handle_array(a), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorPareto(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(a, chunks=chunks)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
from numpy.lib.index_tricks import ndindex

from .. import datasource as _nx


class nd_grid(object):
    """
    Construct a multi-dimensional "meshgrid".

    ``grid = nd_grid()`` creates an instance which will return a mesh-grid
    when indexed.  The dimension and number of the output arrays are equal
    to the number of indexing dimensions.  If the step length is not a
    complex number, then the stop is not inclusive.

    However, if the step length is a **complex number** (e.g. 5j), then the
    integer part of its magnitude is interpreted as specifying the
    number of points to create between the start and stop values, where
    the stop value **is inclusive**.

    If instantiated with an argument of ``sparse=True``, the mesh-grid is
    open (or not fleshed out) so that only one-dimension of each returned
    argument is greater than 1.

    Parameters
    ----------
    sparse : bool, optional
        Whether the grid is sparse or not. Default is False.

    Notes
    -----
    Two instances of `nd_grid` are made available in the Mars.tensor namespace,
    `mgrid` and `ogrid`::

        mgrid = nd_grid(sparse=False)
        ogrid = nd_grid(sparse=True)

    Users should use these pre-defined instances instead of using `nd_grid`
    directly.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mgrid = mt.lib.index_tricks.nd_grid()
    >>> mgrid[0:5,0:5]
    array([[[0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1],
            [2, 2, 2, 2, 2],
            [3, 3, 3, 3, 3],
            [4, 4, 4, 4, 4]],
           [[0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4]]])
    >>> mgrid[-1:1:5j]
    array([-1. , -0.5,  0. ,  0.5,  1. ])

    >>> ogrid = mt.lib.index_tricks.nd_grid(sparse=True)
    >>> ogrid[0:5,0:5]
    [array([[0],
            [1],
            [2],
            [3],
            [4]]), array([[0, 1, 2, 3, 4]])]

    """

    def __init__(self, sparse=False):
        self.sparse = sparse

    def __getitem__(self, key):
        try:
            size = []
            typ = int
            for k in range(len(key)):
                step = key[k].step
                start = key[k].start
                if start is None:
                    start = 0
                if step is None:
                    step = 1
                if isinstance(step, complex):
                    size.append(int(abs(step)))
                    typ = float
                else:
                    size.append(
                        int(math.ceil((key[k].stop - start)/(step*1.0))))
                if (isinstance(step, float) or
                        isinstance(start, float) or
                        isinstance(key[k].stop, float)):
                    typ = float
            if self.sparse:
                nn = [_nx.arange(_x, dtype=_t)
                        for _x, _t in zip(size, (typ,)*len(size))]
            else:
                nn = _nx.indices(size, typ)
            for k in range(len(size)):
                step = key[k].step
                start = key[k].start
                if start is None:
                    start = 0
                if step is None:
                    step = 1
                if isinstance(step, complex):
                    step = int(abs(step))
                    if step != 1:
                        step = (key[k].stop - start)/float(step-1)
                nn[k] = (nn[k]*step+start)
            if self.sparse:
                slobj = [np.newaxis]*len(size)
                for k in range(len(size)):
                    slobj[k] = slice(None, None)
                    nn[k] = nn[k][slobj]
                    slobj[k] = np.newaxis
            return nn
        except (IndexError, TypeError):
            step = key.step
            stop = key.stop
            start = key.start
            if start is None:
                start = 0
            if isinstance(step, complex):
                step = abs(step)
                length = int(step)
                if step != 1:
                    step = (key.stop-start)/float(step-1)
                stop = key.stop + step
                return _nx.arange(0, length, 1, float)*step + start
            else:
                return _nx.arange(start, stop, step)

    def __len__(self):
        return 0


mgrid = nd_grid(sparse=False)
ogrid = nd_grid(sparse=True)


# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import random
import time
import os
from collections import deque, defaultdict
from datetime import datetime

from .assigner import AssignerActor
from .resource import ResourceActor
from .kvstore import KVStoreActor
from .utils import SchedulerActor, remove_shuffle_chunks, GraphState, OperandState
from ..compat import six, OrderedDict
from ..graph import DAG
from ..tiles import handler, DataNotReady
from ..serialize.dataserializer import loads, dumps
from ..utils import serialize_graph, deserialize_graph, merge_tensor_chunks

logger = logging.getLogger(__name__)


class ResultReceiverActor(SchedulerActor):
    def __init__(self):
        super(ResultReceiverActor, self).__init__()
        self._kv_store_ref = None

    @classmethod
    def default_name(cls):
        return 's:%s' % cls.__name__

    def post_create(self):
        self.set_cluster_info_ref()
        self._kv_store_ref = self.get_actor_ref(KVStoreActor.default_name())

    def fetch_tensor(self, session_id, graph_key, tensor_key):
        from ..tensor.expressions.datasource import TensorFetchChunk
        from ..tensor.execution.core import Executor
        from ..worker.transfer import ResultSenderActor

        graph_actor = self.ctx.actor_ref(GraphActor.gen_name(session_id, graph_key))
        fetch_graph = deserialize_graph(graph_actor.build_tensor_merge_graph(tensor_key))

        if len(fetch_graph) == 1 and isinstance(next(fetch_graph.iter_nodes()).op, TensorFetchChunk):
            c = next(fetch_graph.iter_nodes())
            endpoints = self._kv_store_ref.read('/sessions/%s/chunks/%s/workers'
                                                % (session_id, c.key))
            worker_ip = endpoints.children[0].key.rsplit('/', 1)[-1]
            sender_ref = self.ctx.actor_ref(ResultSenderActor.default_name(), address=worker_ip)
            future = sender_ref.fetch_data(session_id, c.key, _wait=False)
            return future.result()
        else:
            ctx = dict()
            target_keys = set()
            for c in fetch_graph:
                if isinstance(c.op, TensorFetchChunk):
                    if c.key in ctx:
                        continue
                    endpoints = self._kv_store_ref.read('/sessions/%s/chunks/%s/workers'
                                                        % (session_id, c.key))
                    worker_ip = endpoints.children[0].key.rsplit('/', 1)[-1]
                    sender_ref = self.ctx.actor_ref(ResultSenderActor.default_name(), address=worker_ip)
                    future = sender_ref.fetch_data(session_id, c.key, _wait=False)
                    ctx[c.key] = future
                else:
                    target_keys.add(c.key)
            ctx = dict((k, loads(future.result())) for k, future in six.iteritems(ctx))
            executor = Executor(storage=ctx)
            concat_result = executor.execute_graph(fetch_graph, keys=target_keys)
            return dumps(concat_result[0])


class GraphActor(SchedulerActor):
    """
    Actor handling execution and status of a Mars graph
    """
    @staticmethod
    def gen_name(session_id, graph_key):
        return 'graph$%s$%s' % (session_id, graph_key)

    def __init__(self, session_id, graph_key, serialized_tensor_graph,
                 target_tensors=None, serialized_chunk_graph=None,
                 state=GraphState.UNSCHEDULED, final_state=None):
        super(GraphActor, self).__init__()
        self._graph_key = graph_key
        self._session_id = session_id
        self._serialized_tensor_graph = serialized_tensor_graph
        self._serialized_chunk_graph = serialized_chunk_graph
        self._state = state
        self._final_state = final_state

        self._start_time = None
        self._end_time = None
        self._nodes_num = None

        self._cluster_info_ref = None
        self._assigner_actor_ref = None
        self._resource_actor_ref = None
        self._kv_store_ref = None

        self._tensor_graph_cache = None
        self._chunk_graph_cache = None
        self._chunk_undigraph_cache = None

        self._op_key_to_chunk = defaultdict(list)
        self._chunk_key_to_chunk = defaultdict(list)

        self._resource_actor = None
        self._tensor_to_tiled = defaultdict(list)
        self._terminal_chunk_op_tensor = defaultdict(set)
        self._terminated_tensors = set()
        self._operand_infos = dict()
        if target_tensors:
            self._target_tensor_chunk_ops = dict((k, set()) for k in target_tensors)
            self._target_tensor_finished = dict((k, set()) for k in self._target_tensor_chunk_ops)
        else:
            self._target_tensor_chunk_ops = dict()
            self._target_tensor_finished = dict()

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        random.seed(int(time.time()))
        self.set_cluster_info_ref()
        self._assigner_actor_ref = self.get_actor_ref(AssignerActor.gen_name(self._session_id))
        self._resource_actor_ref = self.get_actor_ref(ResourceActor.default_name())
        self._kv_store_ref = self.get_actor_ref(KVStoreActor.default_name())

    @property
    def state(self):
        """
        running state of the graph
        """
        return self._state

    @state.setter
    def state(self, value):
        if value != self._state:
            logger.debug('Graph %s state from %s to %s.', self._graph_key, self._state, value)
        self._state = value
        self._kv_store_ref.write(
            '/sessions/%s/graph/%s/state' % (self._session_id, self._graph_key), value.name)

    def __setattr__(self, key, value):
        object.__setattr__(self, key, value)

    @property
    def final_state(self):
        return self._final_state

    @final_state.setter
    def final_state(self, value):
        self._final_state = value
        self._kv_store_ref.write(
            '/sessions/%s/graph/%s/final_state' % (self._session_id, self._graph_key), value.name)

    def execute_graph(self):
        """
        Start graph execution
        """
        self._start_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        self.state = GraphState.PREPARING

        try:
            self.prepare_graph()
            self.scan_node()
            self.place_initial_chunks()
            self.create_operand_actors()
        except:
            logger.exception('Failed to start graph execution.')
            self.stop_graph()
            self.state = GraphState.FAILED
            raise

    def stop_graph(self):
        """
        Stop graph execution
        """
        from .operand import OperandActor
        self.state = GraphState.CANCELLING

        try:
            chunk_graph = self.get_chunk_graph()
        except KeyError:
            self.state = GraphState.CANCELLED
            return

        for chunk in chunk_graph:
            if chunk.op.key not in self._operand_infos:
                continue
            if self._operand_infos[chunk.op.key]['state'] in \
                    (OperandState.READY, OperandState.RUNNING, OperandState.FINISHED):
                # we only need to stop on ready, running and finished operands
                op_uid = OperandActor.gen_uid(self._session_id, chunk.op.key)
                scheduler_addr = self.get_scheduler(op_uid)
                ref = self.ctx.actor_ref(op_uid, address=scheduler_addr)
                ref.stop_operand(_tell=True)

    def reload_chunk_graph(self):
        """
        Reload chunk graph from kv store
        """
        chunk_graph_pb = self._kv_store_ref.read('/sessions/%s/graphs/%s/chunk_graph'
                                                 % (self._session_id, self._graph_key)).value

        self._chunk_graph_cache = DAG.from_pb(chunk_graph_pb)
        self._chunk_undigraph_cache = self._chunk_graph_cache.build_undirected()

        op_key_to_chunk = defaultdict(list)
        chunk_key_to_chunk = defaultdict(list)
        for n in self._chunk_graph_cache:
            chunk_key_to_chunk[n.key].append(n)
            op_key_to_chunk[n.op.key].append(n)
        self._op_key_to_chunk = op_key_to_chunk
        self._chunk_key_to_chunk = chunk_key_to_chunk

    def get_chunk_graph(self):
        if self._chunk_graph_cache is None:
            self.reload_chunk_graph()
        return self._chunk_graph_cache

    def get_chunk_undigraph(self):
        if self._chunk_undigraph_cache is None:
            self.reload_chunk_graph()
        return self._chunk_undigraph_cache

    def prepare_graph(self, compose=True):
        """
        Tile and compose tensor graph into chunk graph
        :param compose: if True, do compose after tiling
        """
        tensor_graph = deserialize_graph(self._serialized_tensor_graph)
        self._tensor_graph_cache = tensor_graph

        logger.debug('Begin preparing graph %s with %d tensors to chunk graph.',
                     self._graph_key, len(tensor_graph))

        # mark target tensor steps
        if not self._target_tensor_chunk_ops:
            for tn in tensor_graph:
                if not tensor_graph.count_successors(tn):
                    self._target_tensor_chunk_ops[tn.key] = set()
                    self._target_tensor_finished[tn.key] = set()

        if self._serialized_chunk_graph:
            serialized_chunk_graph = self._serialized_chunk_graph
            chunk_graph = DAG.from_pb(serialized_chunk_graph)
        else:
            chunk_graph = DAG()

        key_to_chunk = {c.key: c for c in chunk_graph}

        tensor_to_tiled = self._tensor_to_tiled

        for t in tensor_graph:
            if t.key not in tensor_to_tiled:
                continue
            t._chunks = [key_to_chunk[k] for k in [tensor_to_tiled[t.key][-1]]]

        tq = deque()
        for t in tensor_graph:
            if t.inputs and not all(ti.key in tensor_to_tiled for ti in t.inputs):
                continue
            tq.append(t)

        while tq:
            tensor = tq.popleft()
            if not tensor.is_coarse() or tensor.key in tensor_to_tiled:
                continue
            inputs = [tensor_to_tiled[it.key][-1] for it in tensor.inputs or ()]

            op = tensor.op.copy()
            _ = op.new_tensors(inputs, [o.shape for o in tensor.op.outputs],
                               dtype=[o.dtype for o in tensor.op.outputs], **tensor.params)

            total_tiled = []
            for j, t, to_tile in zip(itertools.count(0), tensor.op.outputs, op.outputs):
                # replace inputs with tiled ones
                if not total_tiled:
                    try:
                        td = handler.dispatch(to_tile)
                    except DataNotReady:
                        continue

                    if isinstance(td, (tuple, list)):
                        total_tiled.extend(td)
                    else:
                        total_tiled.append(td)

                tiled = total_tiled[j]
                tensor_to_tiled[t.key].append(tiled)

                # add chunks to fine grained graph
                q = deque([tiled_c.data for tiled_c in tiled.chunks])
                input_chunk_keys = set(itertools.chain(*([(it.key, it.id) for it in input.chunks]
                                                         for input in to_tile.inputs)))
                while len(q) > 0:
                    c = q.popleft()
                    if (c.key, c.id) in input_chunk_keys:
                        continue
                    if c not in chunk_graph:
                        chunk_graph.add_node(c)
                    for ic in c.inputs or []:
                        if ic not in chunk_graph:
                            chunk_graph.add_node(ic)
                            q.append(ic)
                        chunk_graph.add_edge(ic, c)

                for succ in tensor_graph.iter_successors(t):
                    if any(t.key not in tensor_to_tiled for t in succ.inputs):
                        continue
                    tq.append(succ)

        # record the chunk nodes in graph
        reserve_chunk = set()
        result_chunk_keys = list()
        for tk in self._target_tensor_chunk_ops:
            for n in [c.data for t in tensor_to_tiled[tk] for c in t.chunks]:
                result_chunk_keys.append(n.key)
                dq_predecessors = deque([n])
                while dq_predecessors:
                    current = dq_predecessors.popleft()
                    reserve_chunk.add(n)
                    predecessors = chunk_graph.predecessors(current)
                    dq_predecessors.extend([p for p in predecessors if p not in reserve_chunk])
                    reserve_chunk.update(predecessors)
        # delete redundant chunk
        for n in list(chunk_graph.iter_nodes()):
            if n not in reserve_chunk:
               chunk_graph.remove_node(n)

        if compose:
            chunk_graph.compose(keys=result_chunk_keys)

        for tk in tensor_to_tiled:
            if tk not in self._target_tensor_chunk_ops:
                continue
            for n in tensor_to_tiled[tk][-1].chunks:
                self._terminal_chunk_op_tensor[n.op.key].add(tk)
                self._target_tensor_chunk_ops[tk].add(n.op.key)

        # sync chunk graph to kvstore
        graph_path = '/sessions/%s/graphs/%s' % (self._session_id, self._graph_key)
        self._kv_store_ref.write('%s/chunk_graph' % graph_path, serialize_graph(chunk_graph))

        self._nodes_num = len(chunk_graph)
        self._chunk_graph_cache = chunk_graph
        self._chunk_undigraph_cache = self._chunk_graph_cache.build_undirected()
        for n in self._chunk_graph_cache:
            self._chunk_key_to_chunk[n.key].append(n)
            self._op_key_to_chunk[n.op.key].append(n)

    def scan_node(self):
        operand_infos = self._operand_infos
        chunk_graph = self.get_chunk_graph()
        depth_cache = dict()

        for n in chunk_graph.topological_iter():
            key = n.op.key
            if key not in operand_infos:
                operand_infos[key] = dict(
                    optimize=dict(depth=0, demand_depths=(), successor_size=0, descendant_size=0)
                )
            operand_infos[key]['optimize']['successor_size'] += chunk_graph.count_successors(n)
            if not n.inputs:
                depth_cache[key] = 0
            else:
                depth = depth_cache[key] = 1 + max(depth_cache[ni.op.key] for ni in n.inputs)
                # record operand information
                operand_infos[key]['optimize']['depth'] = depth

        for n in chunk_graph.topological_iter(reverse=True):
            operand_infos[n.op.key]['optimize']['descendant_size'] += 1
            if not n.inputs:
                continue
            for ni in set(n.inputs):
                operand_infos[ni.op.key]['optimize']['descendant_size'] += \
                    operand_infos[n.op.key]['optimize']['descendant_size']

    def place_initial_chunks(self):
        """
        Decide target worker for initial chunks
        """
        logger.debug('Placing initial chunks for graph %s', self._graph_key)

        import numpy as np
        graph = self.get_chunk_graph()
        undigraph = self.get_chunk_undigraph()
        operand_infos = self._operand_infos

        metrics = self._resource_actor_ref.get_workers_meta()

        def _successor_fun(n):
            return remove_shuffle_chunks(undigraph.iter_successors(n))

        key_to_chunks = defaultdict(list)
        for n in graph:
            key_to_chunks[n.op.key].append(n)

        # collect chunks with no inputs, or all inputs from shuffle
        zero_degree_key_chunks = dict()
        zero_degrees = []
        for chunk in graph:
            if not remove_shuffle_chunks(graph.predecessors(chunk)) and chunk.op.key not in zero_degree_key_chunks:
                zero_degree_key_chunks[chunk.op.key] = key_to_chunks[chunk.op.key]
        for op_key in zero_degree_key_chunks:
            zero_degrees.append(key_to_chunks[op_key][0])
        random.shuffle(zero_degrees)

        # sort initials by descendant size
        descendant_counts = dict()
        for key, chunks in zero_degree_key_chunks.items():
            descendant_counts[key] = operand_infos[key]['optimize']['descendant_size']
        # note that different sort orders can contribute to different efficiency
        zero_degrees = sorted(zero_degrees, key=lambda n: descendant_counts[n.op.key])

        endpoint_res = [(ep, int(metrics[ep]['hardware']['cpu_total'])) for ep in metrics]
        endpoint_res.sort(key=lambda t: t[1], reverse=True)

        endpoints = [t[0] for t in endpoint_res]
        endpoint_cores = np.array([t[1] for t in endpoint_res])

        splited_initial_counts = (endpoint_cores / endpoint_cores.sum() * len(zero_degrees)).astype(np.int64)
        if splited_initial_counts.sum() < len(zero_degrees):
            pos = 0
            rest = len(zero_degrees) - splited_initial_counts.sum()
            while rest > 0:
                splited_initial_counts[pos] += 1
                rest -= 1
                pos = (pos + 1) % len(splited_initial_counts)

        full_assigns = dict()
        sorted_initials = [v for v in zero_degrees]
        while splited_initial_counts.max():
            slot_id = splited_initial_counts.argmax()
            cur = sorted_initials.pop()
            while cur.op.key in full_assigns:
                cur = sorted_initials.pop()
            assigned = 0
            spread_range = 0
            for v in undigraph.bfs(start=key_to_chunks[cur.op.key], visit_predicate=lambda *_: True,
                                   successors=_successor_fun):
                if v.op.key in full_assigns:
                    continue
                spread_range += 1
                if graph.predecessors(v):
                    continue
                full_assigns[v.op.key] = endpoints[slot_id]
                assigned += 1
                if spread_range >= len(graph) * 1.0 / len(metrics) \
                        or assigned >= splited_initial_counts[slot_id]:
                    break
            splited_initial_counts[slot_id] -= assigned

        # assign initial workers by seeds
        for v in undigraph.bfs(start=zero_degrees, visit_predicate=lambda *_: True, successors=_successor_fun):
            if v.op.key in full_assigns:
                continue
            key_transfers = defaultdict(lambda: 0)
            for pred in undigraph.iter_predecessors(v):
                if pred.op.key in full_assigns:
                    key_transfers[full_assigns[pred.op.key]] += pred.nbytes
            if not key_transfers:
                continue
            max_transfer = max(key_transfers.values())
            max_assigns = [assign for assign, nbytes in key_transfers.items() if nbytes == max_transfer]
            full_assigns[v.op.key] = max_assigns[random.randint(0, len(max_assigns) - 1)]

        for v in zero_degrees:
            operand_infos[v.op.key]['target_worker'] = full_assigns[v.op.key]

    def get_executable_operand_dag(self, op_key, serialize=True):
        """
        Make an operand into a worker-executable dag
        :param op_key: operand key
        :param serialize: whether to return serialized dag
        """
        from ..tensor.expressions.datasource import TensorFetchChunk
        graph = DAG()

        inputs_to_copied = dict()
        for c in self._op_key_to_chunk[op_key]:
            for inp in set(c.inputs or ()):
                op = TensorFetchChunk(dtype=inp.dtype, to_fetch_key=inp.key)
                inp_chunk = op.new_chunk(None, inp.shape, _key=inp.key).data
                inputs_to_copied[inp] = inp_chunk
                graph.add_node(inp_chunk)
            inputs = [inputs_to_copied[inp] for inp in (c.inputs or ())]

            new_op = c.op.copy()
            kws = []
            for o in c.op.outputs:
                kw = {
                    '_key': o.key,
                    'dtype': o.dtype,
                    'index': o.index,
                }
                composed = []
                # copy composed
                for j, com in enumerate(o.composed or []):
                    new_com_op = com.op.copy()
                    if j == 0:
                        inps = inputs
                    else:
                        # if more than 1 inputs, means they are exactly the same object
                        inps = [composed[j - 1]] * len(com.inputs)
                    new_com = new_com_op.new_chunk(inps, com.shape, index=com.index,
                                                   dtype=com.dtype, _key=com.key)
                    composed.append(new_com)
                kw['_composed'] = composed
                kws.append(kw)

            new_outputs = new_op.new_chunks(inputs, [o.shape for o in c.op.outputs],
                                            kws=kws)
            for co in new_outputs:
                exec_chunk = co.data
                graph.add_node(exec_chunk)
                for inp in inputs:
                    graph.add_edge(inp, exec_chunk)
        if serialize:
            return serialize_graph(graph)
        else:
            return graph

    def create_operand_actors(self):
        """
        Create operand actors for all operands
        """
        logger.debug('Creating operand actors for graph %s', self._graph_key)
        from .operand import OperandActor

        chunk_graph = self.get_chunk_graph()
        operand_infos = self._operand_infos

        op_refs = dict()
        write_params = []
        initial_keys = []
        for op_key in self._op_key_to_chunk:
            op_path = '/sessions/%s/operands/%s' % (self._session_id, op_key)

            op_name = type(self._op_key_to_chunk[op_key][0].op).__name__

            # collect operand i/o information
            predecessor_keys = set()
            successor_keys = set()
            input_chunk_keys = set()
            shared_input_chunk_keys = set()
            chunk_key_sizes = dict()

            for c in self._op_key_to_chunk[op_key]:
                predecessor_keys.update(pn.op.key for pn in chunk_graph.iter_predecessors(c))
                successor_keys.update(pn.op.key for pn in chunk_graph.iter_successors(c))
                input_chunk_keys.update(pn.key for pn in chunk_graph.iter_predecessors(c))
                shared_input_chunk_keys.update(pn.key for pn in chunk_graph.iter_predecessors(c)
                                               if chunk_graph.count_successors(pn) > 1)
                chunk_key_sizes.update((co.key, co.nbytes) for co in c.op.outputs)

            io_meta = dict(
                predecessors=list(predecessor_keys),
                successors=list(successor_keys),
                input_chunks=list(input_chunk_keys),
                shared_input_chunks=list(shared_input_chunk_keys),
                chunks=list(chunk_key_sizes.keys()),
            )
            operand_infos[op_key]['op_name'] = op_name
            operand_infos[op_key]['io_meta'] = io_meta
            output_size = sum(chunk_key_sizes.values())
            operand_infos[op_key]['output_size'] = int(output_size)

            if predecessor_keys:
                state = 'UNSCHEDULED'
            else:
                initial_keys.append(op_key)
                state = 'READY'

            write_params.append(('%s/retries' % op_path, 0))
            operand_infos[op_key]['retries'] = 0
            write_params.append(('%s/state' % op_path, 0))
            operand_infos[op_key]['state'] = state

            op_uid = OperandActor.gen_uid(self._session_id, op_key)
            scheduler_addr = self.get_scheduler(op_uid)
            future = self.ctx.create_actor(
                OperandActor, self._session_id, self._graph_key, op_key, operand_infos[op_key],
                is_terminal=op_key in self._terminal_chunk_op_tensor,
                uid=op_uid, address=scheduler_addr,
                wait=False
            )
            op_refs[op_key] = future

            operand_infos[op_key]['state'] = OperandState(state.lower())

        write_future = self._kv_store_ref.write_batch(write_params, _wait=False)
        [future.result() for future in itertools.chain(six.itervalues(op_refs), [write_future])]

        self.state = GraphState.RUNNING

        for op_key in initial_keys:
            op_refs[op_key].result().start_operand(_tell=True)

    def mark_terminal_finished(self, op_key, final_state=None):
        """
        Mark terminal operand as finished. Calling this method will change graph state
        if all terminals are in finished states.
        :param op_key: operand key
        :param final_state: state of the operand
        """
        tensor_keys = self._terminal_chunk_op_tensor[op_key]
        for tensor_key in tensor_keys:
            self._target_tensor_finished[tensor_key].add(op_key)
            if final_state == GraphState.FAILED:
                if self.final_state != GraphState.CANCELLED:
                    self.final_state = GraphState.FAILED
            elif final_state == GraphState.CANCELLED:
                self.final_state = final_state
            if self._target_tensor_finished[tensor_key] == self._target_tensor_chunk_ops[tensor_key]:
                self._terminated_tensors.add(tensor_key)
                if len(self._terminated_tensors) == len(self._target_tensor_chunk_ops):
                    self.state = self.final_state if self.final_state is not None else GraphState.SUCCEEDED
                    self._end_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    def get_state(self):
        return self.state

    def get_graph_info(self):
        return self._start_time, self._end_time, len(self._operand_infos)

    def set_operand_state(self, op_key, state):
        self._operand_infos[op_key]['state'] = OperandState(state)

    def get_operand_info(self):
        return self._operand_infos

    def calc_stats(self):
        states = list(OperandState.__members__.values())
        state_mapping = OrderedDict((v, idx) for idx, v in enumerate(states))
        state_names = [s.name for s in state_mapping]

        op_stats = OrderedDict()
        finished = 0
        total_count = len(self._operand_infos)
        for operand_info in six.itervalues(self._operand_infos):
            op_name = operand_info['op_name']
            state = operand_info['state']
            if state in (OperandState.FINISHED, OperandState.FREED):
                finished += 1
            if op_name not in op_stats:
                op_stats[op_name] = [0] * len(state_mapping)
            stats_list = op_stats[op_name]
            stats_list[state_mapping[state]] += 1

        data_src = OrderedDict([('states', state_names), ])
        for op, state_stats in six.iteritems(op_stats):
            sum_chunks = sum(state_stats)
            data_src[op] = [v * 100.0 / sum_chunks for v in state_stats]

        ops = list(data_src)[1:]
        states = data_src['states']
        transposed = OrderedDict()
        transposed['ops'] = ops
        for sid, state in enumerate(states):
            transposed[state] = list()
            for op in ops:
                transposed[state].append(data_src[op][sid])
        return ops, transposed, finished * 100.0 / total_count

    def free_tensor_data(self, tensor_key):
        from .operand import OperandActor
        tiled_tensor = self._tensor_to_tiled[tensor_key][-1]
        for chunk in tiled_tensor.chunks:
            op_uid = OperandActor.gen_uid(self._session_id, chunk.op.key)
            scheduler_addr = self.get_scheduler(op_uid)
            op_ref = self.ctx.actor_ref(op_uid, address=scheduler_addr)
            op_ref.free_data(_tell=True)

    def build_tensor_merge_graph(self, tensor_key):
        from ..tensor.expressions.merge.concatenate import TensorConcatenate
        from ..tensor.expressions.datasource import TensorFetchChunk

        tiled_tensor = self._tensor_to_tiled[tensor_key][-1]
        graph = DAG()
        if len(tiled_tensor.chunks) == 1:
            # only one chunk, just trigger fetch
            c = tiled_tensor.chunks[0]
            op = TensorFetchChunk(dtype=c.dtype, to_fetch_key=c.key)
            fetch_chunk = op.new_chunk(None, c.shape, index=c.index, _key=c.key).data
            graph.add_node(fetch_chunk)
        else:
            fetch_chunks = []
            for c in tiled_tensor.chunks:
                op = TensorFetchChunk(dtype=c.dtype, to_fetch_key=c.key)
                fetch_chunk = op.new_chunk(None, c.shape, index=c.index, _key=c.key).data
                graph.add_node(fetch_chunk)
                fetch_chunks.append(fetch_chunk)
            chunk = TensorConcatenate(dtype=tiled_tensor.op.dtype).new_chunk(
                fetch_chunks, tiled_tensor.shape).data
            graph.add_node(chunk)
            [graph.add_edge(fetch_chunk, chunk) for fetch_chunk in fetch_chunks]

        return serialize_graph(graph)

    def fetch_tensor_result(self, tensor_key):
        from ..worker.transfer import ResultSenderActor

        # TODO for test
        tiled_tensor = self._tensor_to_tiled[tensor_key][-1]
        if tensor_key not in self._terminated_tensors:
            return None

        ctx = dict()
        for chunk_key in [c.key for c in tiled_tensor.chunks]:
            if chunk_key in ctx:
                continue
            endpoints = self._kv_store_ref.read('/sessions/%s/chunks/%s/workers'
                                                % (self._session_id, chunk_key))
            worker_ip = endpoints.children[0].key.rsplit('/', 1)[-1]
            sender_ref = self.ctx.actor_ref(ResultSenderActor.default_name(), address=worker_ip)
            ctx[chunk_key] = loads(sender_ref.fetch_data(self._session_id, chunk_key))
        return dumps(merge_tensor_chunks(tiled_tensor, ctx))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorBinOp, TensorConstant


class TensorSetImag(operands.SetImag, TensorBinOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorSetImag, self).__init__(_casting=casting, _err=err,
                                            _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return x1.issparse() and x2.issparse()

    @classmethod
    def constant_cls(cls):
        return TensorSetImagConstant


class TensorSetImagConstant(operands.SetImagConstant, TensorConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorSetImagConstant, self).__init__(_casting=casting, _err=err,
                                                    _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, 'issparse') and x1.issparse() and np.isscalar(x2) and x2 == 0:
            return True
        if hasattr(x2, 'issparse') and x2.issparse() and np.isscalar(x1) and x1 == 0:
            return True
        return False


def set_imag(val, imag):
    op = TensorSetImag(dtype=val.dtype)
    return op(val, imag)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ..datasource import tensor as astensor


def flip(m, axis):
    """
    Reverse the order of elements in a tensor along the given axis.

    The shape of the array is preserved, but the elements are reordered.

    Parameters
    ----------
    m : array_like
        Input tensor.
    axis : integer
        Axis in tensor, which entries are reversed.


    Returns
    -------
    out : array_like
        A view of `m` with the entries of axis reversed.  Since a view is
        returned, this operation is done in constant time.

    See Also
    --------
    flipud : Flip a tensor vertically (axis=0).
    fliplr : Flip a tensor horizontally (axis=1).

    Notes
    -----
    flip(m, 0) is equivalent to flipud(m).
    flip(m, 1) is equivalent to fliplr(m).
    flip(m, n) corresponds to ``m[...,::-1,...]`` with ``::-1`` at position n.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> A = mt.arange(8).reshape((2,2,2))
    >>> A.execute()
    array([[[0, 1],
            [2, 3]],

           [[4, 5],
            [6, 7]]])

    >>> mt.flip(A, 0).execute()
    array([[[4, 5],
            [6, 7]],

           [[0, 1],
            [2, 3]]])

    >>> mt.flip(A, 1).execute()
    array([[[2, 3],
            [0, 1]],

           [[6, 7],
            [4, 5]]])

    >>> A = mt.random.randn(3,4,5)
    >>> mt.all(mt.flip(A,2) == A[:,:,::-1,...]).execute()
    True
    """
    m = astensor(m)

    sl = [slice(None)] * m.ndim
    try:
        sl[axis] = slice(None, None, -1)
    except IndexError:
        raise ValueError("axis=%i is invalid for the %i-dimensional input tensor"
                         % (axis, m.ndim))

    return m[tuple(sl)]



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import time
import logging
import os
from collections import defaultdict
from datetime import datetime, timedelta

from .utils import SchedulerActor
from ..config import options


logger = logging.getLogger(__name__)


class ResourceActor(SchedulerActor):
    """
    Actor managing free resources on workers
    """
    def __init__(self):
        super(ResourceActor, self).__init__()
        self._meta_cache = dict()
        self._worker_allocations = dict()

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        super(ResourceActor, self).post_create()
        self.ref().clean_worker()

    @classmethod
    def default_name(cls):
        return 's:%s' % cls.__name__

    def clean_worker(self):
        """
        Remove worker when it does not update its status for a long time
        """
        timeout = options.scheduler.status_timeout
        for worker in list(self._meta_cache.keys()):
            worker_meta = self._meta_cache[worker]
            if 'update_time' not in worker_meta:
                continue

            last_time = datetime.strptime(worker_meta['update_time'], '%Y-%m-%d %H:%M:%S')
            time_delta = timedelta(seconds=timeout)
            if last_time + time_delta < datetime.now():
                del self._meta_cache[worker]

        self.ref().clean_worker(_tell=True, _delay=1)

    def get_workers_meta(self):
        return copy.deepcopy(self._meta_cache)

    def set_worker_meta(self, worker, worker_meta):
        self._meta_cache[worker] = worker_meta

    def allocate_resource(self, session_id, op_key, endpoint, alloc_dict):
        """
        Try allocate resource for operands
        :param session_id: session id
        :param op_key: operand key
        :param endpoint: worker endpoint
        :param alloc_dict: allocation dict, listing resources needed by the operand
        :return: True if allocated successfully
        """
        worker_stats = self._meta_cache[endpoint]['hardware']
        if endpoint not in self._worker_allocations:
            self._worker_allocations[endpoint] = dict()
        if session_id not in self._worker_allocations[endpoint]:
            self._worker_allocations[endpoint][session_id] = dict()
        worker_allocs = self._worker_allocations[endpoint][session_id]

        res_used = defaultdict(lambda: 0)
        free_alloc_keys = []
        for alloc_op_key, alloc in worker_allocs.items():
            d, t = alloc
            for k, v in d.items():
                res_used[k] += v
        for k in free_alloc_keys:
            self.deallocate_resource(session_id, k, endpoint)

        for k, v in alloc_dict.items():
            res_used[k] += v

        sufficient = True
        for k, v in res_used.items():
            if res_used[k] > worker_stats.get(k, 0):
                sufficient = False
                break

        if sufficient:
            self._worker_allocations[endpoint][session_id][op_key] = (alloc_dict, time.time())
            return True
        else:
            return False

    def deallocate_resource(self, session_id, op_key, endpoint):
        """
        Deallocate resource
        :param session_id: session id
        :param op_key: operand key
        :param endpoint: worker endpoint
        """
        try:
            del self._worker_allocations[endpoint][session_id][op_key]
        except KeyError:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin


class TensorRandomSample(operands.RandomSample, TensorRandomOperandMixin):
    __slots__ = '_size',

    def __init__(self, state=None, size=None, dtype=None,
                 gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorRandomSample, self).__init__(_state=state, _size=size,
                                                 _dtype=dtype, _gpu=gpu, **kw)

    def __call__(self, chunks):
        return self.new_tensor(None, None, raw_chunks=chunks)


def random_sample(random_state, size=None, chunks=None, gpu=None, **kw):
    """
    Return random floats in the half-open interval [0.0, 1.0).

    Results are from the "continuous uniform" distribution over the
    stated interval.  To sample :math:`Unif[a, b), b > a` multiply
    the output of `random_sample` by `(b-a)` and add `a`::

      (b - a) * random_sample() + a

    Parameters
    ----------
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : float or Tensor of floats
        Array of random floats of shape `size` (unless ``size=None``, in which
        case a single float is returned).

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.random.random_sample().execute()
    0.47108547995356098
    >>> type(mt.random.random_sample().execute())
    <type 'float'>
    >>> mt.random.random_sample((5,)).execute()
    array([ 0.30220482,  0.86820401,  0.1654503 ,  0.11659149,  0.54323428])

    Three-by-two array of random numbers from [-5, 0):

    >>> (5 * mt.random.random_sample((3, 2)) - 5).execute()
    array([[-3.99149989, -0.52338984],
           [-2.99091858, -0.79479508],
           [-1.23204345, -1.75224494]])
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.dtype('f8')
    size = random_state._handle_size(size)
    op = TensorRandomSample(state=random_state._state, size=size, **kw)
    return op(chunks=chunks)



# Copyright 1999-2017 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from setuptools import setup, find_packages, Extension

import numpy as np
from Cython.Build import cythonize
from Cython.Distutils import build_ext

repo_root = os.path.dirname(os.path.abspath(__file__))

try:
    execfile
except NameError:
    def execfile(fname, globs, locs=None):
        locs = locs or globs
        exec(compile(open(fname).read(), fname, "exec"), globs, locs)

version_file_path = os.path.join(repo_root, 'mars', '_version.py')
version_ns = {'__file__': version_file_path}
execfile(version_file_path, version_ns)

requirements = []
with open(os.path.join(repo_root, 'requirements.txt'), 'r') as f:
    requirements.extend(f.read().splitlines())


extra_requirements=[]
with open(os.path.join(repo_root, 'requirements-extra.txt'), 'r') as f:
    extra_requirements.extend(f.read().splitlines())


long_description = None
if os.path.exists(os.path.join(repo_root, 'README.rst')):
    with open(os.path.join(repo_root, 'README.rst')) as f:
        long_description = f.read()


if os.path.exists(os.path.join(repo_root, '.git')):
    git_info = version_ns['get_git_info']()
    if git_info:
        with open(os.path.join(repo_root, 'mars', '.git-branch'), 'w') as git_file:
            git_file.write('%s %s' % git_info)


if 'CI_MODE' in os.environ:
    for root, dirs, files in os.walk(repo_root):
        for fn in files:
            if not fn.endswith('.pyx'):
                continue
            path = os.path.join(root, fn)
            with open(path, 'rb') as f:
                src = f.read()
            with open(path, 'wb') as f:
                f.write(b'# cython: linetrace=True' + os.linesep.encode('utf-8') + src)


if 'MSC' in sys.version:
    extra_compile_args = ['/Ot', '/I' + os.path.join(repo_root, 'misc')]
    if 'CI_MODE' in os.environ:
        extra_compile_args.extend(['/DCYTHON_TRACE_NOGIL=#1', '/DCYTHON_TRACE=#1'])
    extension_kw = {'extra_compile_args': extra_compile_args}
else:
    extra_compile_args = ['-O3']
    if 'CI_MODE' in os.environ:
        extra_compile_args.extend(['-DCYTHON_TRACE_NOGIL=1', '-DCYTHON_TRACE=1'])
    extension_kw = {'extra_compile_args': extra_compile_args}
extension_kw['include_dirs'] = [np.get_include()]
extensions = [
    Extension('mars.graph', ['mars/graph.pyx'], **extension_kw),
    Extension('mars.utils_c', ['mars/utils_c.pyx'], **extension_kw),
    Extension('mars.lib.gipc', ['mars/lib/gipc.pyx'], **extension_kw),
    Extension('mars.actors.core', ['mars/actors/core.pyx'], **extension_kw),
    Extension('mars.actors.distributor', ['mars/actors/distributor.pyx'], **extension_kw),
    Extension('mars.actors.cluster', ['mars/actors/cluster.pyx'], **extension_kw),
    Extension('mars.actors.pool.messages', ['mars/actors/pool/messages.pyx'], **extension_kw),
    Extension('mars.actors.pool.utils', ['mars/actors/pool/utils.pyx'], **extension_kw),
    Extension('mars.actors.pool.gevent_pool', ['mars/actors/pool/gevent_pool.pyx'], **extension_kw),
    Extension('mars.serialize.core', ['mars/serialize/core.pyx'], **extension_kw),
    Extension('mars.serialize.pbserializer', ['mars/serialize/pbserializer.pyx'], **extension_kw),
    Extension('mars.serialize.jsonserializer', ['mars/serialize/jsonserializer.pyx'], **extension_kw),
]


setup_options = dict(
    name='pymars',
    version=version_ns['__version__'],
    description='MARS: a tensor-based unified framework for large-scale data computation.',
    long_description=long_description,
    author='Qin Xuye',
    author_email='qin@qinxuye.me',
    maintainer='Qin Xuye',
    maintainer_email='qin@qinxuye.me',
    url='http://github.com/mars-project/mars',
    license='Apache License 2.0',
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries',
    ],
    packages=find_packages(exclude=('*.tests.*', '*.tests')),
    include_package_data=True,
    scripts=['scripts/mars-scheduler', 'scripts/mars-worker', 'scripts/mars-web'],
    install_requires=requirements,
    cmdclass={'build_ext': build_ext},
    ext_modules=cythonize(extensions),
    extras_require={'distributed': extra_requirements}
)
setup(**setup_options)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .array import as_same_device, device, is_sparse_module


def _qr(ctx, chunk):
    (a,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        q, r = xp.linalg.qr(a)
        qc, rc = chunk.op.outputs
        ctx[qc.key] = q
        ctx[rc.key] = r


def _svd(ctx, chunk):
    (a,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        u, s, v = xp.linalg.svd(a, full_matrices=False)
        uc, sc, vc = chunk.op.outputs
        ctx[uc.key] = u
        ctx[sc.key] = s
        ctx[vc.key] = v


def _cholesky(ctx, chunk):
    (a,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        if xp is np:
            try:
                import scipy.linalg

                ctx[chunk.key] = scipy.linalg.cholesky(a, lower=chunk.op.lower)
                return
            except ImportError:
                pass

        r = xp.linalg.cholesky(a)
        if not chunk.op.lower:
            r = r.T.conj()

        ctx[chunk.key] = r


def _solve_triangular(ctx, chunk):
    (a, b), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        if xp is not np:
            raise NotImplementedError

        import scipy.linalg
        ctx[chunk.key] = scipy.linalg.solve_triangular(a, b, lower=chunk.op.lower)


def _lu(ctx, chunk):
    import scipy.linalg

    (a,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        p, l, u = scipy.linalg.lu(a)
        pc, lc, uc = chunk.op.outputs
        ctx[pc.key] = p
        ctx[lc.key] = l
        ctx[uc.key] = u


def _norm(ctx, chunk):
    (x,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = xp.linalg.norm(x, ord=chunk.op.ord, axis=chunk.op.axis,
                                        keepdims=chunk.op.keepdims)


def _tensordot(ctx, chunk):
    (a, b), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    axes = chunk.op.a_axes, chunk.op.b_axes
    with device(device_id):
        if not chunk.op.sparse and is_sparse_module(xp):
            # tell sparse to do calculation on numpy or cupy dot
            ctx[chunk.key] = xp.tensordot(a, b, axes, sparse=False)
        else:
            ctx[chunk.key] = xp.tensordot(a, b, axes)


def _dot(ctx, chunk):
    (a, b), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        if not chunk.op.sparse and is_sparse_module(xp):
            # tell sparse to do calculation on numpy or cupy dot
            ctx[chunk.key] = xp.dot(a, b, sparse=False)
        else:
            ctx[chunk.key] = xp.dot(a, b)


def _matmul(ctx, chunk):
    (a, b), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.device, ret_extra=True)

    with device(device_id):
        if not chunk.op.sparse and is_sparse_module(xp):
            # tell sparse to do calculation on numpy or cupy matmul
            ctx[chunk.key] = xp.matmul(a, b, sparse=False)
        else:
            ctx[chunk.key] = xp.matmul(a, b)


def register_linalg_handler():
    from ...operands.arithmetic import TensorDot, Dot, Matmul
    from ...operands.linalg import QR, SVD, Cholesky, SolveTriangular, Norm, LU
    from .core import register

    register(QR, _qr)
    register(SVD, _svd)
    register(Cholesky, _cholesky)
    register(SolveTriangular, _solve_triangular)
    register(LU, _lu)
    register(Norm, _norm)
    register(TensorDot, _tensordot)
    register(Dot, _dot)
    register(Matmul, _matmul)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorLaplace(operands.Laplace, TensorRandomOperandMixin):
    __slots__ = '_loc', '_scale', '_size'
    _input_fields_ = ['_loc', '_scale']

    def __init__(self, state=None, size=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorLaplace, self).__init__(_state=state, _size=size,
                                            _dtype=dtype, _gpu=gpu, **kw)

    def __call__(self, loc, scale, chunks=None):
        return self.new_tensor([loc, scale], None, raw_chunks=chunks)


def laplace(random_state, loc=0.0, scale=1.0, size=None, chunks=None, gpu=None, **kw):
    """
    Draw samples from the Laplace or double exponential distribution with
    specified location (or mean) and scale (decay).

    The Laplace distribution is similar to the Gaussian/normal distribution,
    but is sharper at the peak and has fatter tails. It represents the
    difference between two independent, identically distributed exponential
    random variables.

    Parameters
    ----------
    loc : float or array_like of floats, optional
        The position, :math:`\mu`, of the distribution peak. Default is 0.
    scale : float or array_like of floats, optional
        :math:`\lambda`, the exponential decay. Default is 1.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``loc`` and ``scale`` are both scalars.
        Otherwise, ``np.broadcast(loc, scale).size`` samples are drawn.
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized Laplace distribution.

    Notes
    -----
    It has the probability density function

    .. math:: f(x; \mu, \lambda) = \frac{1}{2\lambda}
                                   \exp\left(-\frac{|x - \mu|}{\lambda}\right).

    The first law of Laplace, from 1774, states that the frequency
    of an error can be expressed as an exponential function of the
    absolute magnitude of the error, which leads to the Laplace
    distribution. For many problems in economics and health
    sciences, this distribution seems to model the data better
    than the standard Gaussian distribution.

    References
    ----------
    .. [1] Abramowitz, M. and Stegun, I. A. (Eds.). "Handbook of
           Mathematical Functions with Formulas, Graphs, and Mathematical
           Tables, 9th printing," New York: Dover, 1972.
    .. [2] Kotz, Samuel, et. al. "The Laplace Distribution and
           Generalizations, " Birkhauser, 2001.
    .. [3] Weisstein, Eric W. "Laplace Distribution."
           From MathWorld--A Wolfram Web Resource.
           http://mathworld.wolfram.com/LaplaceDistribution.html
    .. [4] Wikipedia, "Laplace distribution",
           http://en.wikipedia.org/wiki/Laplace_distribution

    Examples
    --------
    Draw samples from the distribution

    >>> import mars.tensor as mt

    >>> loc, scale = 0., 1.
    >>> s = mt.random.laplace(loc, scale, 1000)

    Display the histogram of the samples, along with
    the probability density function:

    >>> import matplotlib.pyplot as plt
    >>> count, bins, ignored = plt.hist(s.execute(), 30, normed=True)
    >>> x = mt.arange(-8., 8., .01)
    >>> pdf = mt.exp(-abs(x-loc)/scale)/(2.*scale)
    >>> plt.plot(x.execute(), pdf.execute())

    Plot Gaussian for comparison:

    >>> g = (1/(scale * mt.sqrt(2 * np.pi)) *
    ...      mt.exp(-(x - loc)**2 / (2 * scale**2)))
    >>> plt.plot(x.execute(),g.execute())
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().laplace(
            handle_array(loc), handle_array(scale), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorLaplace(state=random_state._state, size=size, gpu=gpu, **kw)
    return op(loc, scale, chunks=chunks)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from .utils import SchedulerActor
from .. import kvstore
from ..config import options

logger = logging.getLogger(__name__)


class _SyncEtcdActor(SchedulerActor):
    """
    Internally hold by KVStoreActor, sync the write option
    """

    def __init__(self):
        super(_SyncEtcdActor, self).__init__()

        self._store = kvstore.get(options.kv_store)

    def read(self, item, recursive=False, sort=False):
        return self._store.read(item, recursive=recursive, sort=sort)

    def write(self, key, value):
        return self._store.write(key, value)

    def write_batch(self, items):
        wrap = lambda x: (x,) if not isinstance(x, tuple) else x
        [self.write(*wrap(it)) for it in items]

    def delete(self, key, dir=False, recursive=False):
        return self._store.delete(key, dir=dir, recursive=recursive)


class KVStoreActor(SchedulerActor):
    def __init__(self, recover=False):
        super(KVStoreActor, self).__init__()

        self._store = kvstore.LocalKVStore()
        self._sync_ref = None

        # if recover, means the KVStoreActor is recovered by some fault tolerance mechanism,
        # at this particular moment, if the item does not exist,
        # we will try to read from etcd
        self._recover = recover

    @classmethod
    def default_name(cls):
        return 's:%s' % cls.__name__

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        if isinstance(kvstore.get(options.kv_store), kvstore.EtcdKVStore):
            self._sync_ref = self.ctx.create_actor(_SyncEtcdActor)

    def pre_destroy(self):
        self.ctx.destroy_actor(self._sync_ref)

    def read(self, item, recursive=False, sort=False, silent=False):
        try:
            # directly read data from the memory kvstore
            return self._store.read(item, recursive=recursive, sort=sort)
        except KeyError:
            if self._recover and self._sync_ref:
                result = self._sync_ref.read(item, recursive=recursive, sort=sort)
                if result.dir:
                    self._store.write(item, result.value)
                else:
                    for child in result.children:
                        self._store.write(child.key, child.value)
                return result
            elif not silent:
                raise

    def read_batch(self, items, silent=False):
        wrap = lambda x: (x,) if not isinstance(x, tuple) else x
        return [self.read(*wrap(it), **dict(silent=silent)) for it in items]

    def write(self, key, value=None, ttl=None, dir=False):
        # write into memory kvstore first
        self._store.write(key, value=value, ttl=ttl, dir=dir)
        # tell _SyncEtcdActor to sync data into etcd asynchronously
        if self._sync_ref:
            self._sync_ref.write(key, value, _tell=True, _wait=False)

    def write_batch(self, items):
        # write into memory kvstore first
        for it in items:
            args = (it,) if not isinstance(it, tuple) else it
            self._store.write(*args)
        # tell _SyncEtcdActor to sync data into etcd asynchronously
        if self._sync_ref:
            self._sync_ref.write_batch(items, _tell=True, _wait=False)

    def delete(self, key, dir=False, recursive=False):
        # delete memory kvstore first
        self._store.delete(key, dir=dir, recursive=recursive)
        # tell _SyncEtcdActor to delete data asynchronously
        if self._sync_ref:
            self._sync_ref.delete(key, dir=dir, recursive=recursive, _tell=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Iterable

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import ValueType, ListField
from .core import TensorNoInput
from .arange import arange
from .empty import empty
from .meshgrid import meshgrid


class TensorIndices(TensorNoInput):
    _op_type_ = OperandDef.TENSOR_INDICES

    _dimensions = ListField('dimensions', ValueType.uint64)

    def __init__(self, dimensions=None, dtype=None, gpu=None, **kw):
        super(TensorIndices, self).__init__(_dimensions=dimensions, _dtype=dtype, _gpu=gpu, **kw)

    @property
    def dimensions(self):
        return self._dimensions


def indices(dimensions, dtype=int, chunks=None):
    """
    Return a tensor representing the indices of a grid.

    Compute a tensor where the subtensors contain index values 0,1,...
    varying only along the corresponding axis.

    Parameters
    ----------
    dimensions : sequence of ints
        The shape of the grid.
    dtype : dtype, optional
        Data type of the result.
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension

    Returns
    -------
    grid : Tensor
        The tensor of grid indices,
        ``grid.shape = (len(dimensions),) + tuple(dimensions)``.

    See Also
    --------
    mgrid, meshgrid

    Notes
    -----
    The output shape is obtained by prepending the number of dimensions
    in front of the tuple of dimensions, i.e. if `dimensions` is a tuple
    ``(r0, ..., rN-1)`` of length ``N``, the output shape is
    ``(N,r0,...,rN-1)``.

    The subtensors ``grid[k]`` contains the N-D array of indices along the
    ``k-th`` axis. Explicitly::

        grid[k,i0,i1,...,iN-1] = ik

    Examples
    --------
    >>> import mars.tensor as mt

    >>> grid = mt.indices((2, 3))
    >>> grid.shape
    (2, 2, 3)
    >>> grid[0].execute()        # row indices
    array([[0, 0, 0],
           [1, 1, 1]])
    >>> grid[1].execute()        # column indices
    array([[0, 1, 2],
           [0, 1, 2]])

    The indices can be used as an index into a tensor.

    >>> x = mt.arange(20).reshape(5, 4)
    >>> row, col = mt.indices((2, 3))
    >>> # x[row, col]  # TODO(jisheng): accomplish this if multiple fancy indexing is supported

    Note that it would be more straightforward in the above example to
    extract the required elements directly with ``x[:2, :3]``.

    """
    from ..merge import stack

    dimensions = tuple(dimensions)
    dtype = np.dtype(dtype)
    raw_chunks = chunks
    if chunks is not None and isinstance(chunks, Iterable):
        chunks = tuple(chunks)
    else:
        chunks = (chunks,) * len(dimensions)

    xi = []
    for ch, dim in zip(chunks, dimensions):
        xi.append(arange(dim, dtype=dtype, chunks=ch))

    grid = None
    if np.prod(dimensions):
        grid = meshgrid(*xi, indexing='ij')

    if grid:
        grid = stack(grid)
    else:
        if raw_chunks is None:
            empty_chunks = None
        else:
            empty_chunks = (1,) + chunks
        grid = empty((len(dimensions),) + dimensions, dtype=dtype, chunks=empty_chunks)

    return grid



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorUnaryOp


class TensorCos(operands.Cos, TensorUnaryOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorCos, self).__init__(_casting=casting, _err=err,
                                        _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x):
        return False


@infer_dtype(np.cos)
def cos(x, out=None, where=None, **kwargs):
    """
    Cosine element-wise.

    Parameters
    ----------
    x : array_like
        Input tensor in radians.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated array is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor
        The corresponding cosine values.

    Notes
    -----
    If `out` is provided, the function writes the result into it,
    and returns a reference to `out`.  (See Examples)

    References
    ----------
    M. Abramowitz and I. A. Stegun, Handbook of Mathematical Functions.
    New York, NY: Dover, 1972.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.cos(mt.array([0, mt.pi/2, mt.pi])).execute()
    array([  1.00000000e+00,   6.12303177e-17,  -1.00000000e+00])
    >>>
    >>> # Example of providing the optional output parameter
    >>> out1 = mt.empty(1)
    >>> out2 = mt.cos([0.1], out1)
    >>> out2 is out1
    True
    >>>
    >>> # Example of ValueError due to provision of shape mis-matched `out`
    >>> mt.cos(mt.zeros((3,3)),mt.zeros((2,2)))
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
    ValueError: operands could not be broadcast together with shapes (3,3) (2,2)
    """
    op = TensorCos(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .array import tensor


def meshgrid(*xi, **kwargs):
    """
    Return coordinate matrices from coordinate vectors.

    Make N-D coordinate arrays for vectorized evaluations of
    N-D scalar/vector fields over N-D grids, given
    one-dimensional coordinate tensors x1, x2,..., xn.

    Parameters
    ----------
    x1, x2,..., xn : array_like
        1-D arrays representing the coordinates of a grid.
    indexing : {'xy', 'ij'}, optional
        Cartesian ('xy', default) or matrix ('ij') indexing of output.
        See Notes for more details.
    sparse : bool, optional
        If True a sparse grid is returned in order to conserve memory.
        Default is False.

    Returns
    -------
    X1, X2,..., XN : Tensor
        For vectors `x1`, `x2`,..., 'xn' with lengths ``Ni=len(xi)`` ,
        return ``(N1, N2, N3,...Nn)`` shaped tensors if indexing='ij'
        or ``(N2, N1, N3,...Nn)`` shaped tensors if indexing='xy'
        with the elements of `xi` repeated to fill the matrix along
        the first dimension for `x1`, the second for `x2` and so on.

    Notes
    -----
    This function supports both indexing conventions through the indexing
    keyword argument.  Giving the string 'ij' returns a meshgrid with
    matrix indexing, while 'xy' returns a meshgrid with Cartesian indexing.
    In the 2-D case with inputs of length M and N, the outputs are of shape
    (N, M) for 'xy' indexing and (M, N) for 'ij' indexing.  In the 3-D case
    with inputs of length M, N and P, outputs are of shape (N, M, P) for
    'xy' indexing and (M, N, P) for 'ij' indexing.  The difference is
    illustrated by the following code snippet::

        xv, yv = mt.meshgrid(x, y, sparse=False, indexing='ij')
        for i in range(nx):
            for j in range(ny):
                # treat xv[i,j], yv[i,j]

        xv, yv = mt.meshgrid(x, y, sparse=False, indexing='xy')
        for i in range(nx):
            for j in range(ny):
                # treat xv[j,i], yv[j,i]

    In the 1-D and 0-D case, the indexing and sparse keywords have no effect.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> nx, ny = (3, 2)
    >>> x = mt.linspace(0, 1, nx)
    >>> y = mt.linspace(0, 1, ny)
    >>> xv, yv = mt.meshgrid(x, y)
    >>> xv.execute()
    array([[ 0. ,  0.5,  1. ],
           [ 0. ,  0.5,  1. ]])
    >>> yv.execute()
    array([[ 0.,  0.,  0.],
           [ 1.,  1.,  1.]])
    >>> xv, yv = mt.meshgrid(x, y, sparse=True)  # make sparse output arrays
    >>> xv.execute()
    array([[ 0. ,  0.5,  1. ]])
    >>> yv.execute()
    array([[ 0.],
           [ 1.]])

    `meshgrid` is very useful to evaluate functions on a grid.

    >>> import matplotlib.pyplot as plt
    >>> x = mt.arange(-5, 5, 0.1)
    >>> y = mt.arange(-5, 5, 0.1)
    >>> xx, yy = mt.meshgrid(x, y, sparse=True)
    >>> z = mt.sin(xx**2 + yy**2) / (xx**2 + yy**2)
    >>> h = plt.contourf(x,y,z)

    """
    from ..base import broadcast_to

    indexing = kwargs.pop('indexing', 'xy')
    sparse = kwargs.pop('sparse', False)

    if kwargs:
        raise TypeError(
            "meshgrid() got an unexpected keyword argument '{0}'".format(list(kwargs)[0]))
    if indexing not in ('xy', 'ij'):
        raise ValueError("Valid values for `indexing` are 'xy' and 'ij'.")

    xi = [tensor(x) for x in xi]
    xi = [a.ravel() for a in xi]
    shape = [x.size for x in xi]

    if indexing == 'xy' and len(xi) > 1:
        xi[0], xi[1] = xi[1], xi[0]
        shape[0], shape[1] = shape[1], shape[0]

    grid = []
    for i, x in enumerate(xi):
        slc = [None] * len(shape)
        slc[i] = slice(None)

        r = x[tuple(slc)]

        if not sparse:
            r = broadcast_to(r, shape)

        grid.append(r)

    if indexing == 'xy' and len(xi) > 1:
        grid[0], grid[1] = grid[1], grid[0]

    return grid



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Iterable
from math import ceil
from numbers import Integral
import operator
import inspect
from functools import wraps

import numpy as np

from ...compat import zip_longest, izip, six, reduce, lkeys


def normalize_shape(shape):
    if isinstance(shape, Iterable):
        return tuple(shape)
    else:
        return shape,


def normalize_chunks(shape, chunks):
    shape = normalize_shape(shape)
    if not isinstance(chunks, tuple):
        if isinstance(chunks, Iterable):
            chunks = tuple(chunks)
        elif isinstance(chunks, six.integer_types):
            chunks = (chunks,) * len(shape)

    if len(shape) != len(chunks):
        raise ValueError('Chunks must have the same dimemsion, '
                         'got shape: %s, chunks: %s' % (shape, chunks))

    chunk_sizes = []
    for size, chunk in izip(shape, chunks):
        if isinstance(chunk, Iterable):
            if not isinstance(chunk, tuple):
                chunk = tuple(chunk)

            if sum(chunk) != size:
                raise ValueError('chunks shape should be of the same length, '
                                 'got shape: %s, chunks: %s' % (size, chunk))
            chunk_sizes.append(chunk)
        else:
            assert isinstance(chunk, six.integer_types)

            sizes = tuple(chunk for _ in range(int(size / chunk))) + \
                    (tuple() if size % chunk == 0 else (size % chunk,))
            chunk_sizes.append(sizes)

    return tuple(chunk_sizes)


def broadcast_shape(*shapes):
    if len(shapes) == 1:
        return shapes[0]

    out_shapes = []
    for ss in zip_longest(*[reversed(s) for s in shapes], fillvalue=-1):
        shape = max(ss)
        if any(i != -1 and i != 1 and i != shape and not np.isnan(i) for i in ss):
            raise ValueError('Operands could not be broadcast together '
                             'with shape {0}'.format(' '.join(map(str, shapes))))
        out_shapes.append(shape)
    return tuple(reversed(out_shapes))


def get_chunk_slices(nsplits, idx):
    return tuple(slice(sum(nsplit[:idx]), sum(nsplit[:idx+1]))
                 for idx, nsplit in zip(idx, nsplits))


def random_state_data(n, random_state=None):
    if not isinstance(random_state, np.random.RandomState):
        random_state = np.random.RandomState(random_state)

    random_data = random_state.bytes(624 * n * 4)  # `n * 624` 32-bit integers
    l = list(np.frombuffer(random_data, dtype=np.uint32).reshape((n, -1)))
    assert len(l) == n
    return l


def validate_axis(ndim, axis):
    if axis >= ndim or axis < -ndim:
        raise ValueError('Axis must be between -%d and %d, got %d' %
                         (ndim, ndim - 1, axis))

    return axis if axis >= 0 else ndim + axis


def inject_dtype(dtype):
    def inner(func):
        @wraps(func)
        def call(*tensors, **kw):
            kw['dtype'] = np.dtype(dtype)
            ret = func(*tensors, **kw)
            if ret is NotImplemented:
                reverse_func = getattr(inspect.getmodule(func), 'r{0}'.format(func.__name__), None)
                if reverse_func is not None:
                    ret = reverse_func(*tensors[::-1], **kw)
                if ret is NotImplemented:
                    raise TypeError(
                        "unsupported operand type(s) for {0}: '{1}' and '{2}".format(
                            func.__name__, *[type(t) for t in tensors]))
            return ret

        return call

    return inner


def infer_dtype(np_func, empty=True, reverse=False, check=True):
    def make_arg(arg):
        if empty:
            return np.empty((1,) * max(1, arg.ndim), dtype=arg.dtype)
        else:
            if hasattr(arg, 'op') and hasattr(arg.op, 'data'):
                arg = arg.op.data
            return arg[(0,) * max(1, arg.ndim)]

    def inner(func):
        @wraps(func)
        def h(*tensors, **kw):
            usr_dtype = np.dtype(kw.pop('dtype')) if 'dtype' in kw else None
            args = [make_arg(t) if hasattr(t, 'ndim') and hasattr(t, 'dtype') else t
                    for t in tensors]
            if reverse:
                args = args[::-1]
            np_kw = dict((k, make_arg(v) if hasattr(v, 'ndim') and hasattr(v, 'dtype') else v)
                         for k, v in six.iteritems(kw) if k != 'out')
            try:
                with np.errstate(all='ignore'):
                    dtype = np_func(*args, **np_kw).dtype
            except:
                dtype = None

            if usr_dtype and dtype:
                if check and not np.can_cast(dtype, usr_dtype):
                    raise TypeError(' No loop matching the specified signature '
                                    'and casting was found for ufunc %s' % np_func)
                kw['dtype'] = usr_dtype
            else:
                kw['dtype'] = dtype

            ret = func(*tensors, **kw)
            if ret is NotImplemented:
                reverse_func = getattr(inspect.getmodule(func), 'r{0}'.format(func.__name__), None) \
                    if not reverse else None
                if reverse_func is not None:
                    ret = reverse_func(*tensors[::-1], **kw)
                if ret is NotImplemented:
                    raise TypeError(
                        "unsupported operand type(s) for {0}: '{1}' and '{2}".format(
                            func.__name__, *[type(t) for t in tensors]))
            return ret

        return h

    return inner


def index_ndim(index):
    from ..core import Tensor

    if isinstance(index, Tensor) and index.dtype == np.bool_:
        # boolean indexing will occupy the ndim
        return index.ndim

    return 1 if index is not None else 0


def replace_ellipsis(index, ndim):
    all_illipsis = list(i for i, idx in enumerate(index) if idx is Ellipsis)
    if len(all_illipsis) > 1:
        raise IndexError("an index can only have a single ellipsis ('...')")
    if not all_illipsis:
        return index

    illipsis_index = all_illipsis[0]
    n_extra = ndim - sum([index_ndim(i) for i in index]) + 1
    return index[:illipsis_index] + (slice(None),) * n_extra + index[illipsis_index+1:]


def calc_sliced_size(size, sliceobj):
    if np.isnan(size):
        return np.nan

    start, stop, step = sliceobj.indices(size)
    return int(ceil(abs((stop - start) / float(step))))


def slice_split(index, sizes):
    size = sum(sizes)

    if isinstance(index, Integral):
        i = 0
        ind = index
        lens = list(sizes)
        while ind >= lens[0]:
            i += 1
            ind -= lens.pop(0)
        return {i: ind}

    assert isinstance(index, slice)
    start, stop, step = index.indices(size)

    slice_all = slice(None)

    if index == slice_all:
        return dict((k, slice_all) for k in range(len(sizes)))

    d = dict()
    if step > 0:
        for i, length in enumerate(sizes):
            if start < length and stop > 0:
                d[i] = slice(start, min(stop, length), step)
                start = (start - length) % step
            else:
                start = start - length
            stop -= length
    else:
        rstart = start  # running start
        chunk_boundaries = np.cumsum(sizes)
        for i, chunk_stop in reversed(list(enumerate(chunk_boundaries))):
            # create a chunk start and stop
            if i == 0:
                chunk_start = 0
            else:
                chunk_start = chunk_boundaries[i - 1]

            # if our slice is in this chunk
            if (chunk_start <= rstart < chunk_stop) and (rstart > stop):
                d[i] = slice(rstart - chunk_stop,
                             max(chunk_start - chunk_stop - 1,
                                 stop - chunk_stop),
                             step)

                # compute the next running start point,
                offset = (rstart - (chunk_start - 1)) % step
                rstart = chunk_start + offset - 1

    # replace 0:20:1 with : if appropriate
    for k, v in d.items():
        if v == slice(0, sizes[k], 1):
            d[k] = slice(None, None, None)

    if not d:  # special case x[:0]
        d[0] = slice(0, 0, 1)

    return d


def is_asc_sorted(arr):
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr)

    if len(arr) == 0:
        return True
    return np.all(arr[:-1] <= arr[1:])


def split_index_into_chunks(chunks, index):
    index = np.asarray(index)
    cum_chunks = np.cumsum(chunks)

    if np.any(index >= cum_chunks[-1]):
        idx = index[index >= cum_chunks[-1]][0]
        err = IndexError('index {0} is out of bounds with size {1}'.format(
            idx, cum_chunks[-1]))
        err.idx = idx
        err.size = cum_chunks[-1]
        raise err

    chunk_idx = np.searchsorted(cum_chunks, index, side='right')

    return [index[chunk_idx == i] - (cum_chunks[i-1] if i > 0 else 0)
            for i in range(len(chunks))]


def decide_unify_split(*splits):
    # TODO (jisheng): In the future, we need more sophisticated way to decide the rechunk split
    # right now, for (2, 2) and (3, 1), we get the rechunk split as (2, 1, 1)
    if not splits:
        return ()
    raw_splits = splits
    splits = set(s for s in splits if len(s) > 1)
    if len(splits) == 1:
        return splits.pop()
    if len(splits) == 0:
        return raw_splits[0]

    if any(np.isnan(sum(s)) for s in splits):
        raise ValueError('Tensor chunk sizes are unknown: {0}'.format(splits))
    if len(set(sum(s) for s in splits)) > 1:
        raise ValueError('Splits do not hava same size: {0}'.format(splits))

    q = [list(s) for s in splits]
    size = sum(q[0])
    cum = 0

    res = []
    while cum < size:
        m = min(s[0] for s in q)
        res.append(m)
        for s in q:
            s[0] -= m
            if s[0] == 0:
                s.pop(0)

        cum += m

    return tuple(res)


def unify_nsplits(*tensor_axes):
    from .rechunk import rechunk

    tensor_splits = [dict((a, split) for a, split in izip(axes, t.nsplits) if split != (1,))
                     for t, axes in tensor_axes]
    common_axes = reduce(operator.and_, [set(lkeys(ts)) for ts in tensor_splits])
    axes_unified_splits = dict((ax, decide_unify_split(*(t[ax] for t in tensor_splits)))
                               for ax in common_axes)

    if len(common_axes) == 0:
        return tuple(t[0] for t in tensor_axes)

    res = []
    for t, axes in tensor_axes:
        new_chunk = dict((i, axes_unified_splits[ax]) for ax, i in zip(axes, range(t.ndim))
                         if ax in axes_unified_splits)
        res.append(rechunk(t, new_chunk).single_tiles())

    return tuple(res)


def unify_chunks(*tensors):
    tensor_axes = [(t, range(t.ndim)) if not isinstance(t, tuple) else t
                   for t in tensors]

    if len(tensor_axes) < 2:
        return tuple(t[0] for t in tensors)

    return unify_nsplits(*tensor_axes)


def check_out_param(out, t, casting):
    from .base import broadcast_to

    if not hasattr(out, 'shape'):
        raise TypeError('return arrays must be a tensor')

    try:
        broadcast_to(t, out.shape)
    except ValueError:
        raise ValueError("operands could not be broadcast together "
                         "with shapes ({0}) ({1})".format(','.join(str(s) for s in t.shape),
                                                          ','.join(str(s) for s in out.shape)))

    if not np.can_cast(t.dtype, out.dtype, casting):
        raise TypeError("output (typecode '{0}') could not be coerced "
                        "to provided output paramter (typecode '{1}') "
                        "according to the casting rule ''{2}''".format(t.dtype.char, out.dtype.char, casting))


def recursive_tile(tensor):
    q = [tensor]
    while q:
        t = q[-1]
        cs = [c for c in t.inputs if c.is_coarse()]
        if cs:
            q.extend(cs)
            continue
        t.single_tiles()
        q.pop()

    return tensor


def decide_chunks(shape, chunks, itemsize):
    from ...config import options

    if chunks is not None:
        if isinstance(chunks, Iterable):
            if not isinstance(chunks, dict):
                chunks = {i: c for i, c in enumerate(chunks)}
        elif isinstance(chunks, six.integer_types):
            chunks = {i: chunks for i in range(len(shape))}
        else:
            raise TypeError('chunks must be iterable, got {0}'.format(type(chunks)))

    if chunks is None:
        chunks = dict()

    nleft = len(shape) - len(chunks)

    if nleft < 0:
        raise ValueError("chunks have more dimensions than input tensor")

    if nleft == 0:
        return normalize_chunks(shape, tuple(chunks[j] for j in range(len(shape))))

    max_chunk_size = options.tensor.chunk_size_limit

    # normalize the dimension which specified first
    dim_to_normalized = {i: normalize_chunks((shape[i],), (c,))[0]
                         for i, c in six.iteritems(chunks)}

    left = {j: [] for j in range(len(shape)) if j not in dim_to_normalized}
    left_unsplit = {j: shape[j] for j in left}
    while True:
        nbytes_occupied = np.prod([max(c) for c in six.itervalues(dim_to_normalized)]) * itemsize
        dim_size = np.maximum(int(np.power(max_chunk_size / nbytes_occupied, 1 / float(len(left)))), 1)
        for j, ns in six.iteritems(left.copy()):
            unsplit = left_unsplit[j]
            ns.append(np.minimum(unsplit, dim_size))
            left_unsplit[j] -= ns[-1]
            if left_unsplit[j] <= 0:
                dim_to_normalized[j] = tuple(ns)
                del left[j]

        if len(left) == 0:
            break

    return tuple(dim_to_normalized[i] for i in range(len(dim_to_normalized)))




# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import math
import contextlib

from .. import kvstore
from ..actors import ActorNotExist
from ..config import options
from ..promise import PromiseActor
from ..cluster_info import HasClusterInfoActor


logger = logging.getLogger(__name__)


class WorkerActor(HasClusterInfoActor, PromiseActor):
    """
    Base class of all worker actors, providing necessary utils
    """
    def __init__(self):
        super(WorkerActor, self).__init__()
        self._callbacks = dict()

        self._kv_store_ref = None
        self._kv_store = None
        if options.kv_store:
            self._kv_store = kvstore.get(options.kv_store)

    @classmethod
    def default_name(cls):
        return 'w:{0}'.format(cls.__name__)

    def post_create(self):
        from ..scheduler.kvstore import KVStoreActor

        logger.debug('Actor %s running in process %d', self.uid, os.getpid())
        try:
            self.set_cluster_info_ref()
            has_cluster_info = True
        except ActorNotExist:
            has_cluster_info = False
        if has_cluster_info:
            self._kv_store_ref = self.get_actor_ref(KVStoreActor.default_name())
        self._init_chunk_store()

    def _init_chunk_store(self):
        import pyarrow.plasma as plasma
        from .chunkstore import PlasmaChunkStore
        self._plasma_client = plasma.connect(options.worker.plasma_socket, '', 0)
        self._chunk_store = PlasmaChunkStore(self._plasma_client, self._kv_store_ref)


class ExpMeanHolder(object):
    """
    Collector of statistics of a given series. The value decays by _factor as time elapses.
    """
    def __init__(self, factor=0.8):
        self._factor = factor
        self._count = 0
        self._v_divided = 0
        self._v_divisor = 0
        self._v2_divided = 0

    def put(self, value):
        self._count += 1
        self._v_divided = self._v_divided * self._factor + value
        self._v_divisor = self._v_divisor * self._factor + 1
        self._v2_divided = self._v2_divided * self._factor + value ** 2

    def count(self):
        return self._count

    def mean(self):
        if self._count == 0:
            return 0
        return self._v_divided * 1.0 / self._v_divisor

    def var(self):
        if self._count == 0:
            return 0
        return self._v2_divided * 1.0 / self._v_divisor - self.mean() ** 2

    def std(self):
        return math.sqrt(self.var())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorOutBinOp


class TensorModf(operands.Modf, TensorOutBinOp):
    def __init__(self, casting='same_kind', dtype=None, sparse=False, **kw):
        super(TensorModf, self).__init__(_casting=casting, _dtype=dtype,
                                         _sparse=sparse, **kw)

    @property
    def _fun(self):
        return np.modf


def modf(x, out1=None, out2=None, out=None, where=None, **kwargs):
    """
    Return the fractional and integral parts of a tensor, element-wise.

    The fractional and integral parts are negative if the given number is
    negative.

    Parameters
    ----------
    x : array_like
        Input tensor.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y1 : Tensor
        Fractional part of `x`.
    y2 : Tensor
        Integral part of `x`.

    Notes
    -----
    For integer input the return values are floats.

    See Also
    --------
    divmod : ``divmod(x, 1)`` is equivalent to ``modf`` with the return values
             switched, except it always has a positive remainder.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.session import new_session

    >>> sess = new_session().as_default()
    >>> sess.run(mt.modf([0, 3.5]))
    (array([ 0. ,  0.5]), array([ 0.,  3.]))
    >>> sess.run(mt.modf(-0.5))
    (-0.5, -0)
    """
    op = TensorModf(**kwargs)
    return op(x, out1=out1, out2=out2, out=out, where=where)
